# -*- coding: utf-8 -*-
"""
The :mod:`PyDynamic.misc.testsignals` module is a collection of test signals
which can be used to simulate dynamic measurements and test methods.

This module contains the following functions:

* :func:`shocklikeGaussian`: signal that resembles a shock excitation as a Gaussian
  followed by a smaller Gaussian of opposite sign
* :func:`GaussianPulse`: Generates a Gaussian pulse at :math:`t_0` with height
  :math:`m_0` and std :math:`sigma`
* :func:`rect`: Rectangular signal of given height and width :math:`t_1 - t_0`
* :func:`squarepulse`: Generates a series of rect functions to represent a square
  pulse signal
* :func:`sine`: Generate a sine signal
"""

import itertools

import numpy as np
from numpy import diff, sqrt, sum, array, corrcoef
from scipy.signal import periodogram
from scipy.special import comb

__all__ = [
    "shocklikeGaussian",
    "GaussianPulse",
    "rect",
    "squarepulse",
    "corr_noise",
    "sine",
]


def shocklikeGaussian(time, t0, m0, sigma, noise=0.0):
    """Generates a signal that resembles a shock excitation as a Gaussian
    followed by a smaller Gaussian of opposite sign.

    Parameters
    ----------
        time : np.ndarray of shape (N,)
            time instants (equidistant)
        t0: float
            time instant of signal maximum
        m0: float
             signal maximum
        sigma: float
             std of main pulse
        noise: float, optional
            std of simulated signal noise

    Returns
    -------
        x: np.ndarray of shape (N,)
            signal amplitudes at time instants

    """

    x = -m0 * (time - t0) / sigma * np.exp(0.5) * np.exp(
        -(time - t0) ** 2 / (2 * sigma ** 2))
    if noise > 0:
        x += np.random.randn(len(time)) * noise
    return x


def GaussianPulse(time, t0, m0, sigma, noise=0.0):
    """Generates a Gaussian pulse at t0 with height m0 and std sigma

    Parameters
    ----------
        time: np.ndarray of shape (N,)
            time instants (equidistant)
        t0 : float
           time instant of signal maximum
        m0 : float
            signal maximum
        sigma : float
            std of pulse
        noise: float, optional
            std of simulated signal noise

    Returns
    -------
        x : np.ndarray of shape (N,)
            signal amplitudes at time instants
    """

    x = m0 * np.exp(-((time - t0) ** 2) / (2 * sigma ** 2))
    if noise > 0:
        x = x + np.random.randn(len(time)) * noise
    return x


def rect(time, t0, t1, height=1, noise=0.0):
    """Rectangular signal of given height and width t1-t0

    Parameters
    ----------
        time : np.ndarray of shape (N,)
            time instants (equidistant)
        t0 : float
            time instant of rect lhs
        t1 : float
            time instant of rect rhs
        height : float
            signal maximum
        noise :float or numpy.ndarray of shape (N,), optional
            float: standard deviation of additive white gaussian noise
            ndarray: user-defined additive noise

    Returns
    -------
        x : np.ndarray of shape (N,)
             signal amplitudes at time instants
    """

    x = np.zeros((len(time),))
    x[np.nonzero(time > t0)] = height
    x[np.nonzero(time > t1)] = 0.0

    # add the noise
    if isinstance(noise, float):
        if noise > 0:
            x = x + np.random.randn(len(time)) * noise
    elif isinstance(noise, np.ndarray):
        if x.size == noise.size:
            x = x + noise
        else:
            raise ValueError("Mismatching sizes of x and noise.")
    else:
        raise NotImplementedError(
            "The given noise is neither of type float nor numpy.ndarray. "
        )

    return x


def squarepulse(time, height, numpulse=4, noise=0.0):
    """Generates a series of rect functions to represent a square pulse signal

    Parameters
    ----------
        time : np.ndarray of shape (N,)
            time instants
        height : float
             height of the rectangular pulses
        numpulse : int
             number of pulses
        noise : float, optional
            std of simulated signal noise

    Returns
    -------
        x : np.ndarray of shape (N,)
            signal amplitude at time instants
    """
    width = (time[-1] - time[0]) / (2 * numpulse + 1)  # width of each individual rect
    x = np.zeros_like(time)
    for k in range(numpulse):
        x += rect(time, (2 * k + 1) * width, (2 * k + 2) * width, height)
    if noise > 0:
        x += np.random.randn(len(time)) * noise
    return x


def sine(time, amp=1.0, freq=2 * np.pi, noise=0.0):
    r""" Generate a sine signal

    Parameters
    ----------
        time : np.ndarray of shape (N,)
            time instants
        amp : float, optional
             amplitude of the sine (default = 1.0)
        freq : float, optional
             frequency of the sine in Hz (default = :math:`2 * \pi`)
        noise : float, optional
            std of simulated signal noise (default = 0.0)

    Returns
    -------
        x : np.ndarray of shape (N,)
            signal amplitude at time instants
    """
    x = amp * np.sin(2 * np.pi / freq * time)
    if noise > 0:
        x += np.random.randn(len(time)) * noise
    return x


def multi_sine(time, amps, freqs, noise=0.0):
    r"""Generate a multi-sine signal as summation of single sine signals

    Parameters
    ----------
        time: np.ndarray of shape (N,)
            time instants
        amps: list or np.ndarray of floating point values
            amplitudes of the sine signals
        freqs: list or np.ndarray of floating point values
            frequencies of the sine signals in Hz
        noise: float, optional
            std of simulated signal noise (default = 0.0)

    Returns
    -------
        x: np.ndarray of shape (N,)
            signal amplitude at time instants
    """

    x = np.zeros_like(time)
    for amp, freq in zip(amps, freqs):
        x += amp * np.sin(freq * time)
    x += np.random.randn(len(x)) * noise ** 2
    return x


class corr_noise(object):
    """Base class for generation of a correlated noise process."""

    def __init__(self, w, sigma, seed=None):
        self.w = w
        self.sigma = sigma
        self.rst = np.random.RandomState(seed)

    def calc_noise(self, N=100):
        z = self.rst.randn(N + 4)
        noise = (
            diff(
                diff(
                    diff(diff(z * self.w ** 4) - 4 * z[1:] * self.w ** 3)
                    + 6 * z[2:] * self.w ** 2
                )
                - 4 * z[3:] * self.w
            )
            + z[4:]
        )
        self.Cw = sqrt(sum([comb(4, l) ** 2 * self.w ** (2 * l) for l in range(5)]))
        self.noise = noise * self.sigma / self.Cw
        return self.noise

    def calc_noise2(self, N=100):
        P = np.ceil(1.5 * N)
        NT = self.rst.randn(P) * self.sigma
        STD = np.zeros(21)
        STD[10] = 1.0
        for _ in itertools.repeat(None, 5):
            NTtmp = NT.copy()
            NT[:-1] = NT[:-1] + self.w * NTtmp[1:]
            NT[-1] = NT[-1] + self.w * NTtmp[-1]
            NT[1:] = NT[1:] + self.w * NTtmp[:-1]
            NT[0] = NT[0] + self.w * NTtmp[-1]
            STDtmp = STD.copy()
            STD[1:] = STD[1:] + self.w * STDtmp[:-1]
            STD[:-1] = STD[:-1] + self.w * STDtmp[1:]
        NT = NT / np.linalg.norm(STD)
        self.noise = NT[:N]
        self.Cw = sqrt(sum([comb(4, l) ** 2 * self.w ** (2 * l) for l in range(5)]))
        return self.noise

    def calc_autocorr(self, lag=10):
        return array(
            [1]
            + [corrcoef(self.noise[:-i], self.noise[i:])[0, 1] for i in range(1, lag)]
        )

    def calc_cov(self):
        def cw(k):
            if np.abs(k) > 4: return 0
            c = sum([comb(4, l) * comb(4, np.abs(k) + l) * self.w ** (
                    np.abs(k) + 2 * l) for l in range(5 - np.abs(k))])
            return c / self.Cw ** 2

        N = len(self.noise)
        Sigma = np.zeros((N, N))
        for m in range(N):
            for n in range(m, N):
                Sigma[m, n] = self.sigma ** 2 * cw(n - m)
        self.Sigma = Sigma + Sigma.T - np.diag(np.diag(Sigma))
        return self.Sigma

    def calc_psd(self, noise=None, Fs=1.0, **kwargs):
        if isinstance(noise, np.ndarray):
            return periodogram(noise, fs=Fs, **kwargs)
        else:
            return periodogram(self.noise, fs=Fs, **kwargs)



# -*- coding: utf-8 -*-
"""
Perform tests on identification sub-packages.

"""

import numpy as np

from PyDynamic.identification import fit_filter
from PyDynamic.misc.SecondOrderSystem import sos_FreqResp


def test_LSIIR():
    # measurement system
    f0 = 36e3  # system resonance frequency in Hz
    S0 = 0.124  # system static gain
    delta = 0.0055  # system damping

    f = np.linspace(0, 80e3, 30)  # frequencies for fitting the system
    Hvals = sos_FreqResp(S0, delta, f0, f)  # frequency response of the 2nd order system

    # %% fitting the IIR filter

    Fs = 500e3  # sampling frequency
    Na = 4
    Nb = 4  # IIR filter order (Na - denominator, Nb - numerator)

    b, a, tau = fit_filter.LSIIR(Hvals, Na, Nb, f, Fs)

    assert len(b) == Nb + 1
    assert len(a) == Na + 1
    assert isinstance(tau, int)



"""
import misc
import identification
import deconvolution
import uncertainty

# See http://mathmet.org/projects/14SIP08 and
# https://www.github.com/eichstaedtPTB/PyDynamic
"""

from .deconvolution.fit_filter import LSFIR_unc, LSFIR_uncMC, LSIIR_unc
from .identification.fit_transfer import fit_sos

from .misc import *
from .model_estimation import *
from .uncertainty import *

__version__ = "1.8.0"

__all__ = [
    "invLSFIR",
    "invLSIIR",
    "invLSFIR_unc",
    "invLSFIR_uncMC",
    "invLSIIR_unc",
    "LSFIR",
    "LSIIR",
    "LSFIR_unc",
    "LSFIR_uncMC",
    "LSIIR_unc",
    "fit_som",
    "fit_sos",
    "FreqResp2RealImag",
    "GUM_DFT",
    "GUM_iDFT",
    "DFT_deconv",
    "DFT_multiply",
    "DFT2AmpPhase",
    "AmpPhase2DFT",
    "AmpPhase2Time",
    "Time2AmpPhase",
    "FIRuncFilter",
    "IIRuncFilter",
    "MC",
    "SMC",
    "UMC",
    "UMC_generic",
    "interp1d_unc",
    "db",
    "grpdelay",
    "mapinside",
    "isstable",
    "kaiser_lowpass",
    "savitzky_golay",
    "impinvar",
    "sos_absphase",
    "sos_realimag",
    "sos_FreqResp",
    "sos_phys2filter",
    "shocklikeGaussian",
    "GaussianPulse",
    "squarepulse",
    "rect",
    "corr_noise",
    "sine",
    "ARMA",
    "print_vec",
    "print_mat",
    "make_semiposdef",
    "make_equidistant",
]



# -*- coding: utf-8 -*-
"""
The :mod:`PyDynamic.uncertainty.interpolate` module implements methods for
the propagation of uncertainties in the application of standard interpolation methods
as provided by :class:`scipy.interpolate.interp1d`.

This module contains the following function:

* :func:`interp1d_unc`: Interpolate arbitrary time series considering the associated
  uncertainties
"""
from typing import Optional, Tuple, Union

import numpy as np
from scipy.interpolate import interp1d

__all__ = ["interp1d_unc", "make_equidistant"]


def interp1d_unc(
    t_new: np.ndarray,
    t: np.ndarray,
    y: np.ndarray,
    uy: np.ndarray,
    kind: Optional[str] = "linear",
    copy=True,
    bounds_error: Optional[bool] = None,
    fill_value: Optional[Union[float, Tuple[float, float], str]] = np.nan,
    fill_unc: Optional[Union[float, Tuple[float, float], str]] = np.nan,
    assume_sorted: Optional[bool] = True,
    returnC: Optional[bool] = False,
) -> Union[
    Tuple[np.ndarray, np.ndarray, np.ndarray],
    Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray],
]:
    r"""Interpolate a 1-D function considering the associated uncertainties

    t and y are arrays of values used to approximate some function :math:`f \colon y
    = f(t)`.

    Note that calling :func:`interp1d_unc` with NaNs present in input values
    results in undefined behaviour.

    An equal number of each of the original timestamps (or frequencies), values and
    associated uncertainties is required.

    Parameters
    ----------
        t_new : (M,) array_like
            A 1-D array of real values representing the timestamps (or frequencies) at
            which to evaluate the interpolated values. t_new can be sorted in any order.
        t : (N,) array_like
            A 1-D array of real values representing timestamps (or frequencies) in
            ascending order.
        y : (N,) array_like
            A 1-D array of real values. The length of y must be equal to the length
            of t.
        uy : (N,) array_like
            A 1-D array of real values representing the standard uncertainties
            associated with y.
        kind : str, optional
            Specifies the kind of interpolation for y as a string ('previous',
            'next', 'nearest' or 'linear'). Default is ‘linear’.
        copy : bool, optional
            If True, the method makes internal copies of t and y. If False,
            references to t and y are used. The default is to copy.
        bounds_error : bool, optional
            If True, a ValueError is raised any time interpolation is attempted on a
            value outside of the range of x (where extrapolation is necessary). If
            False, out of bounds values are assigned fill_value. By default, an error
            is raised unless `fill_value="extrapolate"`.
        fill_value : array-like or (array-like, array_like) or “extrapolate”, optional

            - if a ndarray (or float), this value will be used to fill in for
              requested points outside of the data range. If not provided, then the
              default is NaN.
            - If a two-element tuple, then the first element is used as a fill value
              for `t_new < t[0]` and the second element is used for `t_new > t[-1]`.
              Anything that is not a 2-element tuple (e.g., list or ndarray, regardless
              of shape) is taken to be a single array-like argument meant to be used
              for both bounds as `below, above = fill_value, fill_value`.
            - If “extrapolate”, then points outside the data range will be set
              to the first or last element of the values.

            Both parameters `fill_value` and `fill_unc` should be
            provided to ensure desired behaviour in the extrapolation range.

        fill_unc : array-like or (array-like, array_like) or “extrapolate”, optional
            Usage and behaviour as described in `fill_value` but for the
            uncertainties. Both parameters `fill_value` and `fill_unc` should be
            provided to ensure desired behaviour in the extrapolation range.

        assume_sorted : bool, optional
            If False, values of t can be in any order and they are sorted first. If
            True, t has to be an array of monotonically increasing values.
        returnC : bool, optional
            If True, return sensitivity coefficients for later use. This is only
            available for interpolation kind 'linear' and for
            fill_unc="extrapolate" at the moment. If False sensitivity
            coefficients are not returned and internal computation is
            slightly more efficient.

    If `returnC` is False, which is the default behaviour, the method returns:

    Returns
    -------
        t_new : (M,) array_like
            interpolation timestamps (or frequencies)
        y_new : (M,) array_like
            interpolated values
        uy_new : (M,) array_like
            interpolated associated standard uncertainties

    Otherwise the method returns:

    Returns
    -------
        t_new : (M,) array_like
            interpolation timestamps (or frequencies)
        y_new : (M,) array_like
            interpolated values
        uy_new : (M,) array_like
            interpolated associated standard uncertainties
        C : (M,N) array_like
            sensitivity matrix :math:`C`, which is used to compute the uncertainties
            :math:`U_{y_{new}} = C \cdot \operatorname{diag}(u_y^2) \cdot C^T`

    References
    ----------
        * White [White2017]_
    """
    # This is taken from the class scipy.interpolate.interp1d to copy and sort the
    # arrays in case that is requested and of course extended by the uncertainties.
    # ----------------------------------------------------------------------------------
    t = np.array(t, copy=copy)
    y = np.array(y, copy=copy)
    uy = np.array(uy, copy=copy)

    if not assume_sorted:
        ind = np.argsort(t)
        t = t[ind]
        y = np.take(y, ind)
        uy = np.take(uy, ind)
    # ----------------------------------------------------------------------------------
    # Check for proper dimensions of inputs which are not checked as desired by SciPy.
    if not len(y) == len(uy):
        raise ValueError(
            "Array of associated measurement values' uncertainties must be same length "
            "as array of measurement values."
        )

    # Set up parameter dicts for calls of interp1d. We use it for interpolating the
    # values and "interpolation" (i.e. look-ups) of uncertainties in the trivial cases.
    interp1d_params = {
        "kind": kind,
        "copy": False,
        "bounds_error": bounds_error,
        "assume_sorted": True,
    }

    # Ensure constant values outside the original bounds by setting fill_value
    # and fill_unc accordingly but do not bother in case exception would be thrown.
    if not bounds_error:
        # For the extrapolation to rely on SciPy's handling of fill values in all
        # but the linear case we set those explicitly to boundary values of y and uy
        # respectively but handle those cases separately.
        if fill_value == "extrapolate":
            fill_value = y[0], y[-1]

        if fill_unc == "extrapolate":
            fill_unc = uy[0], uy[-1]
        elif bounds_error is not None and returnC:
            # This means bounds_error is intentionally set to False and we want to
            # extrapolate uncertainties with custom values. Additionally the sensitivity
            # coefficients shall be returned. This is not yet possible, because in this
            # case, we do not know, how to map the provided extrapolation values onto
            # the original values and thus we cannot provide the coefficients. Once we
            # deal with this, we will probably introduce another input parameter
            # fill_sens which is expected to be of shape (N,) or a 2-tuple of this
            # shape, which is then used in C wherever an extrapolation is performed.
            raise NotImplementedError(
                "This feature is not yet implemented.  We are planning to add "
                "another input parameter which is meant to carry the sensitivities "
                "for the extrapolated uncertainties. Get in touch with us, "
                "if you need it to discuss how to proceed."
            )

    # Inter- and extrapolate values in the desired fashion relying on SciPy.
    interp_y = interp1d(t, y, fill_value=fill_value, **interp1d_params)
    y_new = interp_y(t_new)

    if kind in ("previous", "next", "nearest"):
        if returnC:
            raise NotImplementedError(
                "Returning the sensitivity matrix for now is only supported for "
                "interpolation types other than 'previous', 'next' and 'nearest'. Get"
                "in touch with us, if you need this to discuss how to proceed."
            )
        # Look up uncertainties.
        interp_uy = interp1d(t, uy, fill_value=fill_unc, **interp1d_params)
        uy_new = interp_uy(t_new)
    elif kind == "linear":
        # Calculate boolean arrays of indices from t_new which are outside t's bounds...
        extrap_range_below = t_new < np.min(t)
        extrap_range_above = t_new > np.max(t)
        extrap_range = extrap_range_below | extrap_range_above
        # .. and inside t's bounds.
        interp_range = ~extrap_range

        # Initialize the result array for the standard uncertainties.
        uy_new = np.empty_like(y_new)

        # Initialize the sensitivity matrix of shape (M, N) if needed.
        if returnC:
            C = np.zeros((len(t_new), len(uy)), "float64")

        # First extrapolate the according values if required and then
        # compute interpolated uncertainties following White, 2017.

        # If extrapolation is needed, fill in the values provided via fill_unc.
        if np.any(extrap_range):
            # At this point fill_unc is either a float (np.nan is a float as well) or
            # a 2-tuple of floats. In case we have one float we set uy_new to this value
            # inside the extrapolation range.
            if isinstance(fill_unc, float):
                uy_new[extrap_range] = fill_unc
            else:
                # Now fill_unc should be a 2-tuple, which we can fill into uy_new.
                uy_new[extrap_range_below], uy_new[extrap_range_above] = fill_unc

            if returnC:
                # In each row of C corresponding to an extrapolation value below the
                # original range set the first column to 1 and in each row of C
                # corresponding to an extrapolation value above the original range set
                # the last column to 1. It is important to do this before
                # interpolating, because in general those two columns can contain
                # non-zero values in the interpolation range.
                C[:, 0], C[:, -1] = extrap_range_below, extrap_range_above

        # If interpolation is needed, compute uncertainties following White, 2017.
        if np.any(interp_range):
            # This following section is taken mainly from scipy.interpolate.interp1d to
            # determine the indices of the relevant original timestamps (or frequencies)
            # just for the interpolation range.
            # --------------------------------------------------------------------------
            # 2. Find where in the original data, the values to interpolate
            #    would be inserted.
            #    Note: If t_new[n] == t[m], then m is returned by searchsorted.
            t_new_indices = np.searchsorted(t, t_new[interp_range])

            # 3. Clip x_new_indices so that they are within the range of
            #    self.x indices and at least 1.  Removes mis-interpolation
            #    of x_new[n] = x[0]
            t_new_indices = t_new_indices.clip(1, len(t) - 1).astype(int)

            # 4. Calculate the slope of regions that each x_new value falls in.
            lo = t_new_indices - 1
            hi = t_new_indices

            t_lo = t[lo]
            t_hi = t[hi]
            # --------------------------------------------------------------------------
            if returnC:
                # Prepare the sensitivity coefficients, which in the first place
                # inside the interpolation range are the Lagrangian polynomials. We
                # compute the Lagrangian polynomials for all interpolation nodes
                # inside the original range.
                L_1 = (t_new[interp_range] - t_hi) / (t_lo - t_hi)
                L_2 = (t_new[interp_range] - t_lo) / (t_hi - t_lo)

                # Create iterators needed to efficiently fill our sensitivity matrix
                # in the rows corresponding to interpolation range.
                lo_it = iter(lo)
                hi_it = iter(hi)
                L_1_it = iter(L_1)
                L_2_it = iter(L_2)

                # In each row of C set the column with the corresponding
                # index in lo to L_1 and the column with the corresponding
                # index in hi to L_2.
                for index, C_row in enumerate(C):
                    if interp_range[index]:
                        C_row[next(lo_it)] = next(L_1_it)
                        C_row[next(hi_it)] = next(L_2_it)
                # Compute the standard uncertainties avoiding to build the sparse
                # covariance matrix diag(u_y^2). We reduce the equation C diag(u_y^2)
                # C^T for now to a more efficient calculation, which will work as
                # long as we deal with uncorrelated values, so that all information
                # can be found on the diagonal of the covariance and thus the result
                # matrix.
                uy_new[interp_range] = np.sqrt(
                    np.sum(C[interp_range] ** 2 * uy ** 2, 1)
                )
            else:
                # Since we do not need the sensitivity matrix, we compute uncertainties
                # more efficient (although we are actually not so sure about this
                # anymore). The simplification of the equation by pulling out the
                # denominator, just works because we work with the squared Lagrangians.
                # Otherwise we would have to account for the summation order.
                uy_prev_sqr = uy[lo] ** 2
                uy_next_sqr = uy[hi] ** 2
                uy_new[interp_range] = np.sqrt(
                    (t_new[interp_range] - t_hi) ** 2 * uy_prev_sqr
                    + (t_new[interp_range] - t_lo) ** 2 * uy_next_sqr
                ) / (t_hi - t_lo)
    else:
        raise NotImplementedError(
            "%s is unsupported yet. Let us know, that you need it." % kind
        )

    if returnC:
        return t_new, y_new, uy_new, C
    return t_new, y_new, uy_new


def make_equidistant(t, y, uy, dt=5e-2, kind="linear"):
    """ Interpolate non-equidistant time series to equidistant

    Interpolate measurement values and propagate uncertainties accordingly.

    Parameters
    ----------
        t: (N,) array_like
            timestamps (or frequencies)
        y: (N,) array_like
            corresponding measurement values
        uy: (N,) array_like
            corresponding measurement values' standard uncertainties
        dt: float, optional
            desired interval length
        kind: str, optional
            Specifies the kind of interpolation for the measurement values
            as a string ('previous', 'next', 'nearest' or 'linear').

    Returns
    -------
        t_new : (M,) array_like
            interpolation timestamps (or frequencies)
        y_new : (M,) array_like
            interpolated measurement values
        uy_new : (M,) array_like
            interpolated measurement values' standard uncertainties

    References
    ----------
        * White [White2017]_
    """
    # Find t's maximum.
    t_max = np.max(t)

    # Setup new vector of timestamps.
    t_new = np.arange(np.min(t), t_max, dt)

    # Since np.arange in overflow situations results in the biggest values not
    # guaranteed to be smaller than t's maximum', we need to check for this and delete
    # these unexpected values.
    if t_new[-1] > t_max:
        t_new = t_new[t_new <= t_max]

    return interp1d_unc(t_new, t, y, uy, kind)



# -*- coding: utf-8 -*-
"""
    Fit of a FIR filter to a simulated reciprocal frequency response of a
    second-order dynamic system. The deconvolution filter is applied to the
    simulated response of the system to a shock-like Gaussian signal.
    In this example uncertainties associated with the simulated system are
    propagated to the estimate of the input signal.

"""
import numpy as np
import scipy.signal as dsp
import matplotlib.pyplot as plt

import PyDynamic.model_estimation.fit_filter as model_est
from PyDynamic.misc.SecondOrderSystem import sos_phys2filter
from PyDynamic.misc.testsignals import shocklikeGaussian
from PyDynamic.misc.filterstuff import kaiser_lowpass, db
from PyDynamic.uncertainty.propagate_filter import FIRuncFilter
from PyDynamic.misc.tools import make_semiposdef

rst = np.random.RandomState(10)

##### FIR filter parameters
N = 12  # filter order
tau = 6 # time delay

# parameters of simulated measurement
Fs = 500e3      # sampling frequency (in Hz)
Ts = 1 / Fs     # sampling intervall length (in s)

# sensor/measurement system
f0 = 36e3; uf0 = 0.01*f0            # resonance frequency
S0 = 0.124; uS0= 0.001*S0           # static gain
delta = 0.0055; udelta = 0.1*delta  # damping

# transform continuous system to digital filter
bc, ac = sos_phys2filter(S0,delta,f0)   # calculate analogue filter coefficients
b, a = dsp.bilinear(bc, ac, Fs)         # transform to digital filter coefficients

# simulate input and output signals
time = np.arange(0, 4e-3 - Ts, Ts)                              # time values
x = shocklikeGaussian(time, t0 = 2e-3, sigma = 1e-5, m0=0.8)    # input signal
y = dsp.lfilter(b, a, x)                                        # output signal
noise = 1e-3                                                    # noise std
yn = y + rst.randn(np.size(y)) * noise                          # add white noise

# Monte Carlo for calculation of unc. assoc. with [real(H),imag(H)]
runs = 10000
MCf0 = f0 + rst.randn(runs)*uf0                 # MC draws of resonance frequency
MCS0 = S0 + rst.randn(runs)*uS0                 # MC draws of static gain
MCd  = delta+ rst.randn(runs)*udelta            # MC draws of damping
f = np.linspace(0, 120e3, 200)                  # frequencies at which to calculate frequency response
HMC = np.zeros((runs, len(f)),dtype=complex)    # initialise frequency response with zeros
for k in range(runs):       # actual Monte Carlo
    bc_,ac_ = sos_phys2filter(MCS0[k], MCd[k], MCf0[k])     # calculate analogue filter coefficients
    b_,a_ = dsp.bilinear(bc_,ac_,Fs)                        # transform to digital filter coefficients
    HMC[k,:] = dsp.freqz(b_,a_,2*np.pi*f/Fs)[1]             # calculate DFT frequency response

Hc = np.mean(HMC,dtype=complex,axis=0)                      # best estimate (complex-valued)
H = np.r_[np.real(Hc), np.imag(Hc)]                         # best estimate in real, imag
UH= np.cov(np.c_[np.real(HMC),np.imag(HMC)],rowvar=0)       # covariance of real, imag
UH= make_semiposdef(UH, verbose=True)                       # correct for numerical errors

bF, UbF = model_est.invLSFIR_unc(H,UH,N,tau,f,Fs)                             # Calculation of FIR deconvolution filter and its assoc. unc.
CbF = UbF/(np.tile(np.sqrt(np.diag(UbF))[:,np.newaxis],(1,N+1))*
		   np.tile(np.sqrt(np.diag(UbF))[:,np.newaxis].T,(N+1,1)))      # correlation of filter coefficients


# Deconvolution Step1: lowpass filter for noise attenuation
fcut = f0+20e3; low_order = 100                             # cut-off frequency and filter order
blow, lshift = kaiser_lowpass(low_order, fcut, Fs)          # FIR low pass filter coefficients
shift = tau + lshift                                        # delay of low-pass plus that of the FIR deconv filter
# Deconvolution Step2: Application of deconvolution filter
xhat,Uxhat = FIRuncFilter(yn,noise,bF,UbF,shift,blow)       # apply low-pass and FIR deconv filter


# Plot of results
fplot = np.linspace(0, 80e3, 1000)
Hc = dsp.freqz(b,a, 2*np.pi*fplot/Fs)[1]
Hif = dsp.freqz(bF, 1.0, 2 * np.pi * fplot / Fs)[1]
Hl = dsp.freqz(blow, 1.0, 2 * np.pi * fplot / Fs)[1]

plt.figure(1); plt.clf()
plt.plot(fplot*1e-3, db(Hc), fplot*1e-3, db(Hif*Hl), fplot*1e-3, db(Hc*Hif*Hl))
plt.legend(('System freq. resp.', 'compensation filter','compensation result'))
# plt.title('Amplitude of frequency responses')
plt.xlabel('frequency / kHz',fontsize=22)
plt.ylabel('amplitude / dB',fontsize=22)
plt.tick_params(which="both",labelsize=16)

fig = plt.figure(2);plt.clf()
ax = fig.add_subplot(1,1,1)
plt.imshow(UbF,interpolation="nearest")
plt.colorbar(ax=ax)
# plt.title('Uncertainty of deconvolution filter coefficients')

plt.figure(3); plt.clf()
plt.plot(time*1e3,np.c_[x,yn,xhat])
plt.legend(('input signal','output signal','estimate of input'))
# plt.title('time domain signals')
plt.xlabel('time / ms',fontsize=22)
plt.ylabel('signal amplitude / au',fontsize=22)
plt.tick_params(which="both",labelsize=16)
plt.xlim(1.5,4)
plt.ylim(-0.41,0.81)

plt.figure(4);plt.clf()
plt.plot(time*1e3,Uxhat)
# plt.title('Uncertainty of estimated input signal')
plt.xlabel('time / ms',fontsize=22)
plt.ylabel('signal uncertainty / au',fontsize=22)
plt.subplots_adjust(left=0.15,right=0.95)
plt.tick_params(which='both', labelsize=16)
plt.xlim(1.5,4)

plt.figure(5);plt.clf()
plt.errorbar(range(N+1), bF, np.sqrt(np.diag(UbF)))

plt.show()



# -*- coding: utf-8 -*-
"""
The :mod:`PyDynamic.misc.tools` module is a collection of miscellaneous helper
functions.

This module contains the following functions:

* :func:`print_vec`: Print vector (1D array) to the console or return as formatted
  string
* :func:`print_mat`: Print matrix (2D array) to the console or return as formatted
  string
* :func:`make_semiposdef`: Make quadratic matrix positive semi-definite
* :func:`FreqResp2RealImag`: Calculate real and imaginary parts from frequency
  response
* :func:`make_equidistant`: Interpolate non-equidistant time series to equidistant
* :func:`trimOrPad`: trim or pad (with zeros) a vector to desired length
* :func:`progress_bar`: A simple and reusable progress-bar
"""
import sys
from typing import Optional

import numpy as np
from scipy.sparse import eye, issparse
from scipy.sparse.linalg.eigen.arpack import eigs

__all__ = [
    "print_mat",
    "print_vec",
    "make_semiposdef",
    "FreqResp2RealImag",
    "make_equidistant",
    "trimOrPad",
    "progress_bar",
    "shift_uncertainty"
]

def shift_uncertainty(x, ux, shift):
    """Shift the elements in the vector x (and associated uncertainty ux) by shift elements.
        This method uses :class:`numpy.roll` to shift the elements in x and ux. See documentation of np.roll for details.

    Parameters
    ----------
        x: (N,) array_like
            vector of estimates
        ux: float, np.ndarray of shape (N,) or of shape (N,N)
            uncertainty associated with the vector of estimates
        shift: int
            amount of shift

    Returns
    -------
        xs: (N,) array
            shifted vector of estimates
        uxs: float, np.ndarray of shape (N,) or of shape (N,N)
            uncertainty associated with the shifted vector of estimates
    """

    assert(isinstance(shift, int))
    # application of shift to the vector of estimates
    xs = np.roll(x, shift)

    if isinstance(ux, float):       # no shift necessary for ux
        return xs, ux
    if isinstance(ux, np.ndarray):
        if len(ux.shape) == 1:      # uncertainties given as vector
            return xs, np.roll(ux, shift)
        elif len(ux.shape) == 2:      # full covariance matrix
            assert(ux.shape[0]==ux.shape[1])
            uxs = np.roll(ux, (shift, shift), axis=(0,1))
            return xs, uxs
        else:
            raise TypeError("Input uncertainty has incompatible shape")
    else:
        raise TypeError("Input uncertainty has incompatible type")

def trimOrPad(array, length, mode="constant"):
    """Trim or pad (with zeros) a vector to the desired length

    Parameters
    ----------
    array : list, 1D np.ndarray
        original data
    length : int
        length of output
    mode : str, optional
        handed over to np.pad, default "constant"

    Returns
    -------
    array_modified : np.ndarray of shape (length,)
        An array that is either trimmed or zero-padded to achieve
        the required `length`. Both actions are applied to the
        right side of the array
    """
    
    if len(array) < length:  # pad zeros to the right if too short
        return np.pad(array, (0, length - len(array)), mode=mode)
    else:  # trim to given length otherwise
        return array[0:length]


def print_vec(vector, prec=5, retS=False, vertical=False):
    """ Print vector (1D array) to the console or return as formatted string

    Parameters
    ----------
        vector : (M,) array_like
        prec : int
            the precision of the output
        vertical : bool
            print out vertical or not
        retS : bool
            print or return string

    Returns
    -------
        s : str
            if retS is True

    """
    if vertical:
        t = "\n"
    else:
        t = "\t"
    s = "".join(["%1.*g %s" % (int(prec), s, t) for s in vector])
    if retS:
        return s
    else:
        print(s)


def print_mat(matrix, prec=5, vertical=False, retS=False):
    """ Print matrix (2D array) to the console or return as formatted string

    Parameters
    ----------
        matrix : (M,N) array_like
        prec : int
            the precision of the output
        vertical : bool
            print out vertical or not
        retS : bool
            print or return string

    Returns
    -------
        s : str
            if retS is True

    """
    if vertical:
        matrix = matrix.T

    s = "".join(
        [
            print_vec(matrix[k, :], prec=prec, vertical=False, retS=True) + "\n"
            for k in range(matrix.shape[0])
        ]
    )

    if retS:
        return s
    else:
        print(s)


def make_semiposdef(
    matrix: np.ndarray,
    maxiter: Optional[int] = 10,
    tol: Optional[float] = 1e-12,
    verbose: Optional[bool] = False,
) -> np.ndarray:
    """Make quadratic matrix positive semi-definite by increasing its eigenvalues

    Parameters
    ----------
    matrix : array_like of shape (N,N)
        the matrix to process
    maxiter : int, optional
        the maximum number of iterations for increasing the eigenvalues, defaults to 10
    tol : float, optional
        tolerance for deciding if pos. semi-def., defaults to 1e-12
    verbose : bool, optional
        If True print smallest eigenvalue of the resulting matrix, if False (default)
        be quiet

    Returns
    -------
    (N,N) array_like
        quadratic positive semi-definite matrix

    Raises
    ------
    ValueError
        If matrix is not square.
    """
    n, m = matrix.shape
    if n != m:
        raise ValueError("Matrix has to be quadratic")
    # use specialised functions for sparse matrices
    if issparse(matrix):
        # enforce symmetric matrix
        matrix = 0.5 * (matrix + matrix.T)
        # calculate smallest eigenvalue
        e = np.real(eigs(matrix, which="SR", return_eigenvectors=False)).min()
        count = 0
        # increase the eigenvalues until matrix is positive semi-definite
        while e < tol and count < maxiter:
            matrix += (np.absolute(e) + tol) * eye(n, format=matrix.format)
            e = np.real(eigs(matrix, which="SR", return_eigenvectors=False)).min()
            count += 1
        e = np.real(eigs(matrix, which="SR", return_eigenvectors=False)).min()
    # same procedure for non-sparse matrices
    else:
        matrix = 0.5 * (matrix + matrix.T)
        count = 0
        e = np.real(np.linalg.eigvals(matrix)).min()
        while e < tol and count < maxiter:
            e = np.real(np.linalg.eigvals(matrix)).min()
            matrix += (np.absolute(e) + tol) * np.eye(n)
        e = np.real(np.linalg.eigvals(matrix)).min()
    if verbose:
        print("Final result of make_semiposdef: smallest eigenvalue is %e" % e)
    return matrix


def FreqResp2RealImag(Abs, Phase, Unc, MCruns=1e4):
    """ Calculate real and imaginary parts from frequency response

    Calculate real and imaginary parts from amplitude and phase with
    associated uncertainties.

    Parameters
    ----------

        Abs: (N,) array_like
            amplitude values
        Phase: (N,) array_like
            phase values in rad
        Unc: (2N, 2N) or (2N,) array_like
            uncertainties
        MCruns: bool
            Iterations for Monte Carlo simulation

    Returns
    -------

        Re, Im: (N,) array_like
            real and imaginary parts (best estimate)
        URI: (2N, 2N) array_like
            uncertainties assoc. with Re and Im
    """

    if len(Abs) != len(Phase) or 2 * len(Abs) != len(Unc):
        raise ValueError("\nLength of inputs are inconsistent.")

    if len(Unc.shape) == 1:
        Unc = np.diag(Unc)

    Nf = len(Abs)

    AbsPhas = np.random.multivariate_normal(
        np.hstack((Abs, Phase)), Unc, int(MCruns)
    )  # draw MC inputs

    H = AbsPhas[:, :Nf] * np.exp(
        1j * AbsPhas[:, Nf:]
    )  # calculate complex frequency response values
    RI = np.hstack((np.real(H), np.imag(H)))  # transform to real, imag

    Re = np.mean(RI[:, :Nf])
    Im = np.mean(RI[:, Nf:])
    URI = np.cov(RI, rowvar=False)

    return Re, Im, URI


def make_equidistant(*args, **kwargs):
    import warnings

    from ..uncertainty.interpolate import make_equidistant

    warnings.warn(
        "The method :mod:`PyDynamic.misc.tools.make_equidistant` will be moved "
        "to :mod:`PyDynamic.uncertainty.interpolate.make_equidistant` in the next "
        "major release 2.0.0. From version 1.4.3 on you should only use the new method "
        "instead. Please change 'from PyDynamic.misc.tools import make_equidistant' to "
        "'from PyDynamic.uncertainty.interpolate import make_equidistant'.",
        PendingDeprecationWarning,
    )
    return make_equidistant(*args, **kwargs)


def progress_bar(
    count,
    count_max,
    width=30,
    prefix="",
    done_indicator="#",
    todo_indicator=".",
    fout=sys.stdout,
):
    """
    A simple and reusable progress-bar

    Parameters
    ----------
        count: int
            current status of iterations, assumed to be zero-based
        count_max: int
            total number of iterations
        width: int, optional
            width of the actual progressbar (actual printed line will be wider)
        prefix: str, optional
            some text that will be printed in front of the bar (i.e.
            "Progress of ABC:")
        done_indicator: str, optional
            what character is used as "already-done"-indicator
        todo_indicator: str, optional
            what character is used as "not-done-yet"-indicator
        fout: file-object, optional
            where the progress-bar should be written/printed to
    """
    x = int(width * (count + 1) / count_max)
    progressString = "{PREFIX}[{DONE}{NOTDONE}] {COUNT}/{COUNTMAX}\r".format(
        PREFIX=prefix,
        DONE=x * done_indicator,
        NOTDONE=(width - x) * todo_indicator,
        COUNT=count + 1,
        COUNTMAX=count_max,
    )

    fout.write(progressString)



# -*- coding: utf-8 -*-
""" Perform tests on the method *make_equidistant*."""
from typing import Dict, Optional, Tuple, Union

import hypothesis.extra.numpy as hnp
import hypothesis.strategies as st
import numpy as np
from hypothesis import given
from hypothesis.strategies import composite
from pytest import deprecated_call, raises

from PyDynamic.misc.tools import make_equidistant


@composite
def timestamps_values_uncertainties_kind(
    draw,
    min_count: Optional[int] = 2,
    max_count: Optional[int] = None,
    kind_tuple: Optional[Tuple[str]] = ("linear", "previous", "next", "nearest"),
    sorted_timestamps: Optional[bool] = True,
) -> Dict[str, Union[np.ndarray, str]]:
    """Set custom strategy for _hypothesis_ to draw desired input from

    Parameters
    ----------
        draw: callable
            this is a hypothesis internal callable to actually draw from provided
            strategies
        min_count: int
            the minimum number of elements expected inside the arrays of timestamps,
            measurement values and associated uncertainties
        max_count: int
            the maximum number of elements expected inside the arrays of timestamps,
            measurement values and associated uncertainties
        kind_tuple: tuple(str), optional
            the tuple of strings out of "linear", "previous", "next", "nearest",
            "spline", "least-squares" from which the strategy for the
            kind randomly chooses. Defaults to the valid options "linear",
            "previous", "next", "nearest".
        sorted_timestamps: bool
            if the timestamps should be in ascending order or not

    Returns
    -------
        A dict containing the randomly generated expected input parameters t, y, uy,
        dt, kind for make_equidistant()
    """
    # Set the maximum absolute value for floats to be really unique in calculations.
    float_abs_max = 1e64
    # Set all common parameters for timestamps, measurements values and associated
    # uncertainties including allowed ranges and number of elements.
    shape_for_timestamps = hnp.array_shapes(
        max_dims=1, min_side=min_count, max_side=max_count
    )
    strategy_params = {
        "dtype": np.float,
        "shape": shape_for_timestamps,
        "elements": st.floats(
            min_value=-float_abs_max,
            max_value=float_abs_max,
            allow_nan=False,
            allow_infinity=False,
        ),
        "unique": True,
    }
    # Draw "original" timestamps.
    t = draw(hnp.arrays(**strategy_params))
    # Sort timestamps in ascending order.
    if sorted_timestamps:
        t.sort()
    # Reuse "original" timestamps shape for measurements values and associated
    # uncertainties and draw both.
    strategy_params["shape"] = np.shape(t)
    y = draw(hnp.arrays(**strategy_params))
    uy = draw(hnp.arrays(**strategy_params))
    dt = draw(
        st.floats(
            min_value=(np.max(t) - np.min(t)) * 1e-3,
            max_value=(np.max(t) - np.min(t)) / 2,
            exclude_min=True,
            allow_nan=False,
            allow_infinity=False,
        )
    )
    kind = draw(st.sampled_from(kind_tuple))
    return {"t": t, "y": y, "uy": uy, "dt": dt, "kind": kind}


@given(timestamps_values_uncertainties_kind())
def test_deprecated_call_make_equidistant(interp_inputs):
    with deprecated_call():
        make_equidistant(**interp_inputs)


# noinspection PyArgumentList
@given(timestamps_values_uncertainties_kind())
def test_too_short_call_make_equidistant(interp_inputs):
    # Check erroneous calls with too few inputs.
    with raises(TypeError):
        make_equidistant(interp_inputs["t"])
        make_equidistant(interp_inputs["t"], interp_inputs["y"])


@given(timestamps_values_uncertainties_kind())
def test_full_call_make_equidistant(interp_inputs):
    t_new, y_new, uy_new = make_equidistant(**interp_inputs)
    # Check the equal dimensions of the minimum calls output.
    assert len(t_new) == len(y_new) == len(uy_new)


@given(timestamps_values_uncertainties_kind())
def test_wrong_input_lengths_call_make_equidistant(interp_inputs):
    # Check erroneous calls with unequally long inputs.
    with raises(ValueError):
        y_wrong = np.tile(interp_inputs["y"], 2)
        uy_wrong = np.tile(interp_inputs["uy"], 3)
        make_equidistant(interp_inputs["t"], y_wrong, uy_wrong)


@given(timestamps_values_uncertainties_kind())
def test_t_new_to_dt_make_equidistant(interp_inputs):
    t_new = make_equidistant(**interp_inputs)[0]
    delta_t_new = np.diff(t_new)
    # Check if the new timestamps are ascending.
    assert not np.any(delta_t_new < 0)


@given(timestamps_values_uncertainties_kind(kind_tuple=("previous", "next", "nearest")))
def test_prev_in_make_equidistant(interp_inputs):
    y_new, uy_new = make_equidistant(**interp_inputs)[1:3]
    # Check if all 'interpolated' values are present in the actual values.
    assert np.all(np.isin(y_new, interp_inputs["y"]))
    assert np.all(np.isin(uy_new, interp_inputs["uy"]))


@given(timestamps_values_uncertainties_kind(kind_tuple=["linear"]))
def test_linear_in_make_equidistant(interp_inputs):
    y_new, uy_new = make_equidistant(**interp_inputs)[1:3]
    # Check if all interpolated values lie in the range of the original values.
    assert np.all(np.amin(interp_inputs["y"]) <= y_new)
    assert np.all(np.amax(interp_inputs["y"]) >= y_new)


@given(st.integers(min_value=3, max_value=1000))
def test_linear_uy_in_make_equidistant(n):
    # Check for given input, if interpolated uncertainties equal 1 and
    # :math:`sqrt(2) / 2`.
    dt_unit = 2
    t_unit = np.arange(0, n, dt_unit)
    y = uy_unit = np.ones_like(t_unit)
    dt_half = dt_unit / 2
    uy_new = make_equidistant(t_unit, y, uy_unit, dt_half, "linear")[2]
    assert np.all(uy_new[0:n:dt_unit] == 1) and np.all(
        uy_new[1:n:dt_unit] == np.sqrt(2) / 2
    )


@given(timestamps_values_uncertainties_kind(kind_tuple=("spline", "least-squares")))
def test_raise_not_implemented_yet_make_equidistant(interp_inputs):
    # Check that not implemented versions raise exceptions.
    with raises(NotImplementedError):
        make_equidistant(**interp_inputs)



# -*- coding: utf-8 -*-
"""Collection of noise-signals

This module contains the following functions:

* :func:`get_alpha`: normal distributed signal amplitudes with equal power
  spectral density
* :func:`power_law_noise`: normal distributed signal amplitudes with power spectrum
  `:math:f^\alpha`
* :func:`power_law_acf`: (theoretical) autocorrelation function of power law noise
* :func:`ARMA`: autoregressive moving average noise process
"""

import numpy as np
from scipy.linalg import toeplitz

__all__ = ["get_alpha", "white_gaussian", "power_law_noise", "power_law_acf", "ARMA"]


# define an alpha for every color
colors = {"violet": 2, "blue": 1, "white": 0, "pink": -1, "red": -2, "brown": -2}


def get_alpha(color_value=0):
    """
    Translate a color (given as string) into an exponent alpha or directly
    hand through a given numeric value of alpha.

    Parameters
    ----------
        color_value: str, int or float
            if string -> check against known colornames -> return alpha
            if numeric -> directly return value

    Returns
    -------
        alpha: float
    """

    if isinstance(color_value, str):
        if color_value in colors.keys():
            alpha = colors[color_value]
        else:
            raise NotImplementedError(
            "Specified color ({COLOR}) of noise is not available. " \
            "Please choose from {COLORS} or define alpha directly." \
            .format(COLOR=color_value, COLORS='/'.join(colors.keys())))

    elif isinstance(color_value, (float, int)):
        alpha = color_value

    else:
        raise IOError("No valid string or numeric value for alpha given")

    return float(alpha)


def white_gaussian(N, mean=0, std=1):
    return np.random.normal(loc=mean, scale=std, size=N)


def power_law_noise(N=None, w=None, color_value="white", mean=0.0, std=1.0):
    """
    Generate colored noise by
    * generate white gaussian noise
    * multiplying its Fourier-transform with f^(alpha/2)
    * inverse Fourier-transform to yield the colored/correlated noise
    * further adjustments to fit to specified mean/std

    based on [Zhivomirov2018](A Method for Colored Noise Generation)

    Parameters
    ----------
        N: int
            length of noise to be generated
        w: numpy.ndarray
            user-defined white noise
            if provided, `N` is ignored!
        color_value: str, int or float
            if string -> check against known colornames
            if numeric -> used as alpha to shape PSD
        mean: float
            mean of the output signal
        std: float
            standard deviation of the output signal

    Returns
    -------
        w_filt: filtered noise signal
    """

    if (N is not None) and (w is not None):
        raise UserWarning("You specified N and w. Ignoring N.")

    # draw white gaussian noise, or take the provided w
    if isinstance(w, np.ndarray):
        N = len(w)
    else:
        w = white_gaussian(N)

    # get alpha either directly or from color-string
    alpha = get_alpha(color_value)

    # (real) fourier transform to get spectrum
    W = np.fft.rfft(w)

    # get index of frequencies
    # note:
    # * this gives [1., 2., 3., ..., N+1] (in accordance with [Zhivomirov2018])
    # * ==> not W_filt ~ f^alpha, but rather W_filt ~ k^alpha
    steps = N // 2 + 1
    k = np.linspace(0, steps, steps) + 1

    # generate the filtered spectrum by multiplication with f^(alpha/2)
    W_filt = W * np.power(k, alpha / 2)

    # calculate the filtered time-series (inverse fourier of modified spectrum)
    w_filt = np.fft.irfft(W_filt, N)

    # adjust to given mean + std
    w_filt = (w_filt - np.mean(w_filt)) / np.std(w_filt)
    w_filt = mean + std * w_filt

    return w_filt


def power_law_acf(N, color_value="white", std=1.0):
    """
    Return the theoretic right-sided autocorrelation (Rww) of different colors of noise. 

    Colors of noise are defined to have a power spectral density (Sww) proportional to `f^\alpha`.
    Sww and Rww form a Fourier-pair. Therefore Rww = ifft(Sww).
    """
    # process the arguments
    alpha = get_alpha(color_value)

    # get index of frequencies (see notes of same line at power_law_noise() )
    k = np.linspace(0, N, N) + 1

    # generate and transform the power spectral density Sww
    Sww = np.power(k, alpha)

    # normalize Sww to have the same overall power as the white-noise-PSD it is transformed from
    #Sww = Sww / np.sum(Sww) * len(k)   # probably unnecessary because of later normalization of Rww

    # inverse Fourier-transform to get Autocorrelation from PSD/Sww
    Rww = np.fft.irfft(Sww)
    Rww = std**2 * Rww / Rww[0]           # This normalization ensures the given standard-deviation

    return Rww[:N]


def ARMA(length, phi=0.0, theta=0.0, std=1.0):
    r"""
    Generate time-series of a predefined ARMA-process based on this equation:
    :math:`\sum_{j=1}^{\min(p,n-1)} \phi_j \epsilon[n-j] + \sum_{j=1}^{\min(q,n-1)} \theta_j w[n-j]`
    where w is white gaussian noise. Equation and algorithm taken from [Eichst2012]_ .

    Parameters
    ----------
    length: int
        how long the drawn sample will be
    phi: float, list or numpy.ndarray, shape (p, )
        AR-coefficients
    theta: float, list or numpy.ndarray
        MA-coefficients
    std: float
        std of the gaussian white noise that is feeded into the ARMA-model

    Returns
    -------
    e: numpy.ndarray, shape (length, )
       time-series of the predefined ARMA-process

    References
    ----------
        * Eichstädt, Link, Harris and Elster [Eichst2012]_
    """

    # convert to numpy.ndarray
    if isinstance(phi, (float, int)):
        phi = np.array([phi])
    elif isinstance(phi, list):
        phi = np.array(phi)

    if isinstance(theta, (float, int)):
        theta = np.array([theta])
    elif isinstance(theta, list):
        theta = np.array(theta)

    # initialize e, w
    w = np.random.normal(loc=0, scale=std, size=length)
    e = np.zeros_like(w)

    # define shortcuts
    p = len(phi)
    q = len(theta)

    # iterate series over time
    for n, wn in enumerate(w):
        min_pn = min(p, n)
        min_qn = min(q, n)
        e[n] = np.sum(phi[:min_pn].dot(e[n-min_pn:n])) + np.sum(theta[:min_qn].dot(w[n-min_qn:n])) + wn

    return e



# -*- coding: utf-8 -*-
"""
.. deprecated:: 1.2.71
    The package *identification* will be combined with the package *deconvolution* and
    renamed to *model_estimation* in the next major release 2.0.0. From version 1.4.1 on
    you should only use the new package :doc:`PyDynamic.model_estimation` instead.

The module :mod:`PyDynamic.identification.fit_transfer` contains several functions
for the identification of transfer function models.

This module for now still contains the following function:

* :func:`fit_sos`: Fit second-order model to complex-valued frequency response

"""
import warnings

from ..model_estimation.fit_transfer import fit_som

__all__ = ["fit_sos"]

warnings.simplefilter("default")
warnings.warn(
    "The package *identification* will be combined with the package "
    "*deconvolution* and renamed to *model_estimation* in the "
    "next major release 2.0.0. From version 1.4.1 on you should only use "
    "the new package *model_estimation* instead.",
    DeprecationWarning,
)


def fit_sos(f, H, UH=None, weighting=None, MCruns=None, scaling=1e-3):
    """.. deprecated:: 1.4.1 Please use :func:`PyDynamic.model_estimation.fit_som`"""
    return fit_som(f=f, H=H, UH=UH, weighting=weighting, MCruns=MCruns, scaling=scaling)



# -*- coding: utf-8 -*-
"""

Python transcript of Octave impinvar function for application of impulse invariance

Note that this module is covered by the GPLv3 license.

"""

import numpy as np
import scipy.signal as dsp
from math import factorial
from scipy.special import binom

polyrev = lambda p_in: p_in[::-1]

def prepad(vec, length, value=0):
	if length > len(vec):
		return np.r_[value * np.ones(length - len(vec)), vec]
	elif length < len(vec):
		return vec[-length:]
	else:
		return vec

def h1_z_deriv(n, p, ts):
	"""
	## Copyright (C) 2007 R.G.H. Eschauzier <reschauzier@yahoo.com>
	## Copyright (C) 2016 S. Eichstädt <sascha.eichstaedt@ptb.de>
	##
	## This program is free software; you can redistribute it and/or modify it under
	## the terms of the GNU General Public License as published by the Free Software
	## Foundation; either version 3 of the License, or (at your option) any later
	## version.
	##
	## This program is distributed in the hope that it will be useful, but WITHOUT
	## ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	## FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
	## details.
	##
	## You should have received a copy of the GNU General Public License along with
	## this program; if not, see <http://www.gnu.org/licenses/>.

	## -*- texinfo -*-
	## @deftypefn {Function File} {@var{b} =} h1_z_deriv (@var{n}, @var{p}, @var{ts})
	## Undocumented internal function.  This function is used by the impinvar
	## and invimpinvar functions in the signal package.
	## @end deftypefn

	## Adapted by Carnë Draug on 2011 <carandraug+dev@gmail.com>

	## Find {-zd/dz}^n*H1(z). I.e., first differentiate, then multiply by -z, then differentiate, etc.
	## The result is (ts^(n+1))*(b(1)*p/(z-p)^1 + b(2)*p^2/(z-p)^2 + b(n+1)*p^(n+1)/(z-p)^(n+1)).
	## Works for n>0.
	"""
	# Build the vector d that holds coefficients for all the derivatives of H1(z)
	## The results reads d(n)*z^(1)*(d/dz)^(1)*H1(z) + d(n-1)*z^(2)*(d/dz)^(2)*H1(z) +...+ d(1)*z^(n)*(d/dz)^(n)*H1(z)

	d = [(-1)**n] # Vector with the derivatives of H1(z)
	for i in range(n-1):
		d  = [d, 0]                     # Shift result right (multiply by -z)
		d += prepad(np.polyder(d), i + 1, 0) # Add the derivative

	## Build output vector
	b = np.zeros(n + 1)
	for i in range(n):
		b += d[i] * prepad(h1_deriv(n-i+1), n+1, 0)

	b *= ts**(n+1)/factorial(n)

	## Multiply coefficients with p^i, where i is the index of the coeff.
	b *=p**(np.arange(n+1,0,-1))

	return b



def h1_deriv(n):
	"""
	Find (z^n)*(d/dz)^n*H1(z), where H1(z)=ts*z/(z-p), ts=sampling period,
	p=exp(sm*ts) and sm is the s-domain pole with multiplicity n+1.
	The result is (ts^(n+1))*(b(1)*p/(z-p)^1 + b(2)*p^2/(z-p)^2 + b(n+1)*p^(n+1)/(z-p)^(n+1)),
	where b(i) is the binomial coefficient bincoeff(n,i) times n!. Works for n>0.
	"""
	bincoeffs = np.zeros(n+1)
	for k in range(n+1):
		bincoeffs[k] = binom(n, k)
	b  = factorial(n)*bincoeffs # Binomial coefficients: [1], [1 1], [1 2 1], [1 3 3 1], etc.
	b *= (-1)**n
	return b



def z_res(r_in, sm, ts):
	p_out = np.exp(ts * sm)  # z-domain pole
	n = len(r_in)  # Multiplicity of the pole
	r_out = np.zeros(n, dtype = complex)  # Residue vector

	## First pole (no multiplicity)
	k_out = to_real(r_in[0] * ts)  # PFE of z/(z-p) = p/(z-p)+1; direct part
	r_out[0] = r_in[0] * ts * p_out  # pole part of PFE

	for i in range(1, n):  # Go through s-domain residues for multiple pole
		r_out[:i] += r_in[i] * polyrev(h1_z_deriv(i, p_out, ts))  # Add z-domain residues

	return r_out, p_out, k_out

def to_real(p_in):
	## Copyright (C) 2007 R.G.H. Eschauzier <reschauzier@yahoo.com>
	## Copyright (C) 2016 S. Eichstädt <sascha.eichstaedt@ptb.de>
	##
	## This program is free software; you can redistribute it and/or modify it under
	## the terms of the GNU General Public License as published by the Free Software
	## Foundation; either version 3 of the License, or (at your option) any later
	## version.
	##
	## This program is distributed in the hope that it will be useful, but WITHOUT
	## ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	## FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
	## details.
	##
	## You should have received a copy of the GNU General Public License along with
	## this program; if not, see <http://www.gnu.org/licenses/>.

	## -*- texinfo -*-
	## @deftypefn {Function File} {@var{p_out} =} to_real (@var{p_in})
	## Undocumented internal function.  This function is used by the impinvar
	## and invimpinvar functions in the signal package.
	## @end deftypefn

	## Adapted by Carnë Draug on 2011 <carandraug+dev@gmail.com>

	## Round complex number to nearest real number

	return np.abs(p_in)*np.sign(np.real(p_in))



def impinvar(b_in, a_in, fs = 1, tol = 0.0001):
	"""

	## Copyright (c) 2007 R.G.H. Eschauzier <reschauzier@yahoo.com>
	## Copyright (c) 2011 Carnë Draug <carandraug+dev@gmail.com>
	## Copyright (c) 2011 Juan Pablo Carbajal <carbajal@ifi.uzh.ch>
	## Copyright (c) 2016 Sascha Eichstaedt <sascha.eichstaedt@ptb.de>
	##
	## This program is free software; you can redistribute it and/or modify it under
	## the terms of the GNU General Public License as published by the Free Software
	## Foundation; either version 3 of the License, or (at your option) any later
	## version.
	##
	## This program is distributed in the hope that it will be useful, but WITHOUT
	## ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	## FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
	## details.
	##
	## You should have received a copy of the GNU General Public License along with
	## this program; if not, see <http://www.gnu.org/licenses/>.
	"""

	ts = 1 / fs  # we should be using sampling frequencies to be compatible with Matlab

	[r_in, p_in, k_in] = dsp.residue(b_in, a_in)  # partial fraction expansion

	n = len(r_in)  # Number of poles/residues

	if (len(k_in) > 0):  # Greater than zero means we cannot do impulse invariance
		if abs(k_in)>0:
			raise ValueError("Order numerator >= order denominator")

	r_out = np.zeros(n, dtype = complex)  # Residues of H(z)
	p_out = np.zeros(n, dtype = complex)  # Poles of H(z)
	k_out = np.zeros(1)  # Constant term of H(z)

	i = 0
	while (i < n):
		m = 0
		first_pole = p_in[i]  # Pole in the s-domain
		while (i < (n-1) and np.abs(first_pole - p_in[i + 1]) < tol):  # Multiple poles at p(i)
			i += 1 # Next residue
			m += 1  # Next multiplicity
		[r, p, k] = z_res(r_in[i - m:i+1], first_pole, ts) # Find z-domain residues
		k_out += k  # Add direct term to output
		p_out[i - m:i+1]   = p  # Copy z-domain pole(s) to output
		r_out[i - m:i+1]   = r  # Copy z-domain residue(s) to output

		i += 1  # Next s-domain residue/pole

	[b_out, a_out] = dsp.invres(r_out, p_out, k_out, tol=tol)
	a_out = to_real(a_out)  # Get rid of spurious imaginary part
	b_out = to_real(b_out)

	## Shift results right to account for calculating in z instead of z^-1
	b_out = b_out[:-1]

	return b_out, a_out

def impinvar_causal(*args, **kwargs):
	"""causal version of impinvar, h[0] == 0"""

	# polynomial.polysub is ordered from lowest to highest
	from numpy.polynomial.polynomial import polysub

	bz, az = impinvar(*args, **kwargs)

	h0 = bz[0] / az[0]
	bz = polysub(bz, h0 * az)
	return bz, az



# -*- coding: utf-8 -*-
"""
The :doc:`PyDynamic.deconvolution` package implements methods for the design of
digital deconvolution filters by least-squares fitting to the reciprocal of a given
frequency response with associated uncertainties.
"""

# See https://mathmet.org/projects/14SIP08 and
# https://www.github.com/PTB-PSt1/PyDynamic

from .fit_filter import (
    LSFIR,
    LSIIR,
    LSFIR_unc,
    LSFIR_uncMC,
    LSIIR_unc,
)

__all__ = [
    "LSFIR",
    "LSIIR",
    "LSFIR_unc",
    "LSFIR_uncMC",
    "LSIIR_unc",
]



# -*- coding: utf-8 -*-
"""
The `PyDynamic.misc` module provides various functions and methods which are
used in the examples and in some of the other implemented routines.

# See http://mathmet.org/projects/14SIP08 and
# https://www.github.com/eichstaedtPTB/PyDynamic
"""

__all__ = [
    "db",
    "grpdelay",
    "mapinside",
    "isstable",
    "kaiser_lowpass",
    "savitzky_golay",
    "impinvar",
    "sos_absphase",
    "sos_realimag",
    "sos_FreqResp",
    "sos_phys2filter",
    "shocklikeGaussian",
    "GaussianPulse",
    "squarepulse",
    "rect",
    "corr_noise",
    "sine",
    "print_vec",
    "print_mat",
    "make_semiposdef",
    "make_equidistant",
    "FreqResp2RealImag",
    "ARMA",
]

from .SecondOrderSystem import sos_FreqResp, sos_phys2filter, sos_absphase, sos_realimag
from .filterstuff import (
    db,
    grpdelay,
    mapinside,
    isstable,
    kaiser_lowpass,
    savitzky_golay,
)
from .impinvar import impinvar
from .testsignals import (
    shocklikeGaussian,
    GaussianPulse,
    squarepulse,
    rect,
    corr_noise,
    sine,
)
from .noise import ARMA
from .tools import (
    print_mat,
    print_vec,
    make_semiposdef,
    FreqResp2RealImag,
    make_equidistant,
)



# -*- coding: utf-8 -*-
"""
Created on Thu Sep 10 11:58:46 2015

In this example an actually measured signal is deconvolved and the result compared to a reference measurement signal.

"""
import os

import numpy as np
import matplotlib.pyplot as plt
from PyDynamic.uncertainty.propagate_DFT import (
    GUM_DFT,
    GUM_iDFT,
    DFT_deconv,
    AmpPhase2DFT,
    DFT_multiply,
)


def demonstrate_deconvolution():
    file_directory = os.path.dirname(os.path.abspath(__file__))
    colors = [[0.1, 0.6, 0.5], [0.9, 0.2, 0.5], [0.9, 0.5, 0.1]]

    # low-pass filter for deconvolution
    def lowpass(f, fcut=80e6):
        return 1 / (1 + 1j * f / fcut) ** 2

    #%% reference data
    ref_file = np.loadtxt(
        os.path.join(file_directory, "DFTdeconv reference_signal.dat")
    )
    time = ref_file[:, 0]
    ref_data = ref_file[:, 1]
    Ts = 2e-9
    N = len(time)

    #%% measured hydrophone output signal
    meas = np.loadtxt(
        os.path.join(file_directory, "DFTdeconv measured_signal.dat")
    )
    y = meas[:, 1]
    # assumed noise std
    noise_std = 4e-4
    Uy = noise_std ** 2

    #%% hydrophone calibration data
    calib = np.loadtxt(
        os.path.join(file_directory, "DFTdeconv calibration.dat")
    )
    f = calib[:, 0]
    FR = calib[:, 1] * np.exp(1j * calib[:, 3])
    Nf = 2 * (len(f) - 1)

    uAmp = calib[:, 2]
    uPhas = calib[:, 4]
    UAP = np.r_[uAmp, uPhas * np.pi / 180] ** 2

    print("Propagating uncertainty associated with measurement through DFT")
    Y, UY = GUM_DFT(y, Uy, N=Nf)
    # propagation to real/imag
    print(
        "Propagating uncertainty associated with calibration data to real and imag part"
    )
    H, UH = AmpPhase2DFT(np.abs(FR), np.angle(FR), UAP)
    # propagation through deconvolution equation
    print("Propagating uncertainty through the inverse system")
    XH, UXH = DFT_deconv(H, Y, UH, UY)
    #%% low pass filter for deconvolution regularization
    HLc = lowpass(f)
    HL = np.r_[np.real(HLc), np.imag(HLc)]
    # application of low-pass filter
    print("Propagating uncertainty through the low-pass filter")
    XH, UXH = DFT_multiply(XH, HL, UXH)
    # propagation back to time domain
    print("Propagating uncertainty associated with the estimate back to time domain\n")
    xh, Uxh = GUM_iDFT(XH, UXH, Nx=N)
    ux = np.sqrt(np.diag(Uxh))

    #%% plotting results
    plt.figure(1)
    plt.clf()
    plt.plot(
        time * 1e6, xh, label="estimated pressure signal", linewidth=2, color=colors[0]
    )
    plt.plot(
        time * 1e6, ref_data, "--", label="reference data", linewidth=2, color=colors[1]
    )
    plt.fill_between(time * 1e6, xh + 2 * ux, xh - 2 * ux, alpha=0.2, color=colors[0])
    plt.xlabel(r"time / $\mu$s", fontsize=22)
    plt.ylabel("signal amplitude / MPa", fontsize=22)
    plt.tick_params(which="major", labelsize=18)
    plt.legend(loc="best", fontsize=18, fancybox=True)
    plt.xlim(0, 2)

    dB = lambda a: 20 * np.log10(np.abs(a))
    plt.figure(2)
    plt.clf()
    plt.plot(f * 1e-6, dB(FR), label="measurement system", linewidth=2)
    plt.plot(f * 1e-6, dB(HLc / FR), label="inverse system", linewidth=2)
    plt.plot(f * 1e-6, dB(HLc), "--", linewidth=0.5)
    plt.legend(loc="best", fontsize=18)
    plt.xlim(0.5, 100)
    plt.ylim(-40, 40)
    plt.xlabel("frequency / MHz", fontsize=22)
    plt.ylabel("frequency response amplitude / dB", fontsize=22)
    plt.tick_params(which="both", labelsize=18)

    plt.figure(3)
    plt.clf()
    plt.subplot(211)
    plt.errorbar(
        f * 1e-6,
        np.abs(FR),
        2 * np.sqrt(UAP[: len(UAP) // 2]),
        fmt=".-",
        alpha=0.2,
        color=colors[0],
    )
    plt.xlim(0.5, 80)
    plt.ylim(0.04, 0.24)
    plt.xlabel("frequency / MHz", fontsize=22)
    plt.tick_params(which="both", labelsize=18)
    plt.ylabel("amplitude / V/MPa", fontsize=22)
    plt.subplot(212)
    plt.errorbar(
        f * 1e-6,
        np.unwrap(np.angle(FR)) * np.pi / 180,
        2 * UAP[len(UAP) // 2 :],
        fmt=".-",
        alpha=0.2,
        color=colors[0],
    )
    plt.xlim(0.5, 80)
    plt.ylim(-0.2, 0.3)
    plt.xlabel("frequency / MHz", fontsize=22)
    plt.tick_params(which="both", labelsize=18)
    plt.ylabel("phase / rad", fontsize=22)

    plt.figure(4)
    plt.clf()
    plt.plot(time * 1e6, ux, label="uncertainty", linewidth=2, color=colors[0])
    plt.xlabel(r"time / $\mu$s", fontsize=22)
    plt.ylabel("uncertainty / MPa", fontsize=22)
    plt.tick_params(which="major", labelsize=18)
    plt.xlim(0, 2)

    plt.show()


if __name__ == "__main__":
    demonstrate_deconvolution()



from matplotlib import pyplot

from examples.uncertainty_for_dft.deconv_DFT import demonstrate_deconvolution


def test_deconvolution_example(monkeypatch):
    # Test executability of the deconvolution example.
    # With this expression we override the matplotlib.pyplot.show method with a
    # lambda expression returning None but only for this one test.
    monkeypatch.setattr(pyplot, "show", lambda: None, raising=True)
    demonstrate_deconvolution()



"""
Perform validation for uncertainty.propagate_filter with (MC) methods
"""
import sys
sys.path.append(".")

import matplotlib.pyplot as plt
import numpy as np
from scipy.linalg import toeplitz

from PyDynamic.misc.testsignals import rect
from PyDynamic.misc.tools import make_semiposdef, trimOrPad
from PyDynamic.misc.filterstuff import kaiser_lowpass
from PyDynamic.misc.noise import power_law_acf, power_law_noise, white_gaussian
from PyDynamic.uncertainty.propagate_MonteCarlo import MC
from PyDynamic.uncertainty.propagate_filter import FIRuncFilter


# parameters of simulated measurement
Fs = 100e3        # sampling frequency (in Hz)
Ts = 1 / Fs       # sampling interval length (in s)

# nominal system parameters
fcut = 20e3                            # low-pass filter cut-off frequency (6 dB)
L = 100                                 # filter order
b1 = kaiser_lowpass(L,   fcut,Fs)[0]
b2 = kaiser_lowpass(L-20,fcut,Fs)[0]

# uncertain knowledge: cutoff between 19.5kHz and 20.5kHz
runs = 1000
FC = fcut + (2*np.random.rand(runs)-1)*0.5e3

B = np.zeros((runs,L+1))
for k in range(runs):        # Monte Carlo for filter coefficients of low-pass filter
    B[k,:] = kaiser_lowpass(L,FC[k],Fs)[0]

Ub = make_semiposdef(np.cov(B,rowvar=0))    # covariance matrix of MC result

# simulate input and output signals
nTime = 500
time  = np.arange(nTime)*Ts                     # time values

# different cases
sigma_noise = 1e-2                              # 1e-5

for kind in ["float", "corr", "diag"]:

    for blow in [None, b2]:

        print(kind, type(blow))

        if kind == "float":
            # input signal + run methods
            x = rect(time,100*Ts,250*Ts,1.0,noise=sigma_noise)

            y, Uy = FIRuncFilter(x, sigma_noise, b1, Ub, blow=blow, kind=kind)       # apply uncertain FIR filter (GUM formula)
            yMC, UyMC = MC(x,sigma_noise,b1,[1.0],Ub,runs=runs,blow=blow)             # apply uncertain FIR filter (Monte Carlo)
            
        elif kind == "corr":

            # get an instance of noise, the covariance and the covariance-matrix with the specified color
            color = "red"
            noise = power_law_noise(N=nTime, color_value=color, std=sigma_noise)

            Ux = power_law_acf(nTime, color_value=color, std=sigma_noise)

            # input signal
            x = rect(time,100*Ts,250*Ts,1.0,noise=noise)

            # build Ux_matrix from autocorrelation Ux
            Ux_matrix = toeplitz(trimOrPad(Ux, nTime))

            # run methods
            y, Uy = FIRuncFilter(x, Ux, b1, Ub, blow=blow, kind=kind)              # apply uncertain FIR filter (GUM formula)
            yMC, UyMC = MC(x, Ux_matrix, b1, [1.0], Ub, runs=runs, blow=blow)             # apply uncertain FIR filter (Monte Carlo)

        elif kind == "diag":
            sigma_diag = sigma_noise * ( 1 + np.heaviside(np.arange(len(time)) - len(time)//2.5,0) )    # std doubles after half of the time
            noise = sigma_diag * white_gaussian(len(time))

            # input signal + run methods
            x = rect(time,100*Ts,250*Ts,1.0,noise=noise)

            y, Uy = FIRuncFilter(x, sigma_diag, b1, Ub, blow=blow, kind=kind)            # apply uncertain FIR filter (GUM formula)
            yMC, UyMC = MC(x, sigma_diag, b1, [1.0], Ub, runs=runs, blow=blow)             # apply uncertain FIR filter (Monte Carlo)

        # compare FIR and MC results
        plt.figure(1); plt.cla()
        plt.plot(time, x, label="input")
        plt.plot(time, y, label="output FIR direct")
        plt.plot(time, yMC, label="output FIR MC")
        plt.xlabel("time [s]")
        plt.ylabel("signal amplitude [1]")
        plt.legend()

        plt.figure(2);plt.cla()
        plt.plot(time, Uy, label="FIR formula")
        plt.plot(time, np.sqrt(np.diag(UyMC)), label="Monte Carlo")
        plt.xlabel("time [s]")
        plt.ylabel("signal uncertainty [1]")
        plt.legend()
        plt.show()


# -*- coding: utf-8 -*-
"""
The :mod:`PyDynamic.misc.SecondOrderSystem` module is a collection of methods that
are used throughout the whole package, specialized for second
order dynamic systems, such as the ones used for high-class accelerometers.

This module contains the following functions:

* :func:`sos_FreqResp`: Calculation of the system frequency response
* :func:`sos_phys2filter`: Calculation of continuous filter coefficients from
  physical parameters
* :func:`sos_absphase`: Propagation of uncertainty from physical parameters to real
  and imaginary	part of system's transfer function using GUM S2 Monte Carlo
* :func:`sos_realimag`: Propagation of uncertainty from physical parameters to real
  and imaginary	part of system's transfer function using GUM S2 Monte Carlo

"""

__all__ = ["sos_FreqResp", "sos_phys2filter", "sos_absphase", "sos_realimag"]

import numpy as np

ua = lambda a: np.unwrap(np.angle(a))


def sos_FreqResp(S, d, f0, freqs):
    """Calculation of the system frequency response

    The frequency response is calculated from the continuous physical model
    of a second order system given by

    :math:`H(f) = \\frac{4S\\pi^2f_0^2}{(2\\pi f_0)^2 + 2jd(2\\pi f_0)f - f^2}`

    If the provided system parameters are vectors then :math:`H(f)` is calculated for
    each set of parameters. This is helpful for Monte Carlo simulations by using
    draws from the model parameters

    Parameters
    ----------
        S:      float or ndarray shape (K,)
                static gain
        d:      float or ndarray shape (K,)
                damping parameter
        f0:     float or ndarray shape (K,)
                resonance frequency
        freqs:  ndarray shape (N,)
                frequencies at which to calculate the freq response

    Returns
    -------
        H:  ndarray shape (N,) or ndarray shape (N,K)
            complex frequency response values

    """
    om0 = 2 * np.pi * f0
    rho = S * (om0 ** 2)
    w = 2 * np.pi * freqs

    if isinstance(S, np.ndarray):
        H = np.tile(rho, (len(w), 1)) * (
            om0 ** 2
            + 2j
            * np.tile(d * om0, (len(w), 1))
            * np.tile(w[:, np.newaxis], (1, len(S)))
            - np.tile(w[:, np.newaxis] ** 2, (1, len(S)))
        ) ** (-1)
    else:
        H = rho / (om0 ** 2 + 2j * d * om0 * w - w ** 2)

    return H


def sos_phys2filter(S, d, f0):
    """Calculation of continuous filter coefficients from physical parameters.

    If the provided system parameters are vectors then the filter coefficients
    are calculated for each set of parameters. This is helpful for Monte Carlo
    simulations by using draws from the model parameters

    Parameters
    ----------
        S:  float
            static gain
        d:  float
            damping parameter
        f0: float
            resonance frequency

    Returns
    -------
        b,a: ndarray
            analogue filter coefficients
    """

    if isinstance(S, np.ndarray):
        bc = [S * (2 * np.pi * f0) ** 2]
        ac = np.c_[np.ones((len(S),)), 4 * d * np.pi * f0, (2 * np.pi * f0) ** 2]
    else:
        bc = [S * (2 * np.pi * f0) ** 2]
        ac = np.array([1, 2 * d * 2 * np.pi * f0, (2 * np.pi * f0) ** 2])

    return bc, ac


def sos_realimag(S, d, f0, uS, ud, uf0, f, runs=10000):
    """Propagation of uncertainty from physical parameters to real and imaginary
    part of system's transfer function using GUM S2 Monte Carlo.

    Parameters
    ----------
        S:    float
            static gain
        d:    float
            damping
        f0:   float
            resonance frequency
        uS:   float
            uncertainty associated with static gain
        ud:   float
            uncertainty associated with damping
        uf0:  float
            uncertainty associated with resonance frequency
        f:    ndarray, shape (N,)
            frequency values at which to calculate real and imaginary part

    Returns
    -------
        Hmean:   ndarray, shape (N,)
            best estimate of complex frequency response values
        Hcov:    ndarray, shape (2N,2N)
            covariance matrix [ [u(real,real), u(real,imag)], [u(imag,real), u(imag,imag)] ]
    """

    runs = int(runs)
    SMC = S + np.random.randn(runs) * uS
    dMC = d + np.random.randn(runs) * ud
    fMC = f0 + np.random.randn(runs) * uf0

    HMC = sos_FreqResp(SMC, dMC, fMC, f)

    return (
        np.mean(HMC, dtype=complex, axis=1),
        np.cov(np.vstack((np.real(HMC), np.imag(HMC))), rowvar=1),
    )


def sos_absphase(S, d, f0, uS, ud, uf0, f, runs=10000):
    """Propagation of uncertainty from physical parameters to real and imaginary
    part of system's transfer function using GUM S2 Monte Carlo.

    Parameters
    ----------
        S:    float
            static gain
        d:    float
            damping
        f0:   float
            resonance frequency
        uS:   float
            uncertainty associated with static gain
        ud:   float
            uncertainty associated with damping
        uf0:  float
            uncertainty associated with resonance frequency
        f:    ndarray, shape (N,)
            frequency values at which to calculate amplitue and phase

    Returns
    -------
        Hmean:   ndarray, shape (N,)
            best estimate of complex frequency response values
        Hcov:    ndarray, shape (2N,2N)
            covariance matrix [ [u(abs,abs), u(abs,phase)], [u(phase,abs), u(phase,phase)] ]
    """

    runs = int(runs)
    SMC = S + np.random.randn(runs) * uS
    dMC = d + np.random.randn(runs) * ud
    fMC = f0 + np.random.randn(runs) * uf0

    HMC = sos_FreqResp(SMC, dMC, fMC, f)

    return (
        np.mean(HMC, dtype=complex, axis=1),
        np.cov(np.vstack((np.abs(HMC), ua(HMC))), rowvar=1),
    )



# -*- coding: utf-8 -*-

"""

Uncertainty propagation for an IIR lowpass filter with uncertain cut-off frequency for a rectangular signal.

"""

import matplotlib.pyplot as plt
import scipy.signal as dsp
import numpy as np

from PyDynamic.misc.testsignals import rect
from PyDynamic.uncertainty.propagate_filter import IIRuncFilter
import PyDynamic.uncertainty.propagate_MonteCarlo as MC
from PyDynamic.misc.tools import make_semiposdef

# parameters of simulated measurement
Fs = 100e3		# sampling frequency (in Hz)
Ts = 1.0/Fs		# sampling interval length (in s)

# nominal system parameter
fcut = 20e3		# low pass filter cut-off frequency (in Hz)
L = 6			# filter order
b,a = dsp.butter(L,2*fcut/Fs,btype='lowpass')	# nominal filter coefficients

# uncertain knowledge: fcut between 19.8kHz and 20.2kHz
runs = 1000		# number of Monte Carlo runs
FC = fcut + (2*np.random.rand(runs)-1)*0.2e3	# cut-off frequency range
AB = np.zeros((runs,len(b)+len(a)-1))			# filter coefficients

for k in range(runs):		# Monte Carlo for uncertainty of filter coefficients
	bb,aa = dsp.butter(L,2*FC[k]/Fs,btype='lowpass')
	AB[k,:] = np.hstack((aa[1:],bb))

Uab = make_semiposdef(np.cov(AB,rowvar=0))	# correct for numerical errors

# Apply filter
time = np.arange(0,499*Ts,Ts)				# time values
t0 = 100*Ts; t1 = 300*Ts					# left and right boundary of rect signal
height = 0.9								# input signal height
noise = 1e-3								# std of white noise
x = rect(time,t0,t1,height,noise=noise)		# input signal

y,Uy = IIRuncFilter(x, noise, b, a, Uab)		# apply IIR formula (GUM)
yMC,UyMC = MC.SMC(x,noise,b,a,Uab,runs=10000)	# apply sequential Monte Carlo method (GUM S1)

plt.figure(1);plt.cla()
plt.plot(time*1e3, x, label="input signal")
plt.plot(time*1e3, y, label="output signal")
plt.xlabel('time / ms',fontsize=22)
plt.ylabel('signal amplitude / au',fontsize=22)
plt.tick_params(which="both",labelsize=16)

plt.figure(2);plt.cla()
plt.plot(time*1e3, Uy, label='IIR formula')
plt.plot(time*1e3, UyMC, label='Monte Carlo')
# plt.title('uncertainty of filter output')
plt.legend()
plt.xlabel('time / ms',fontsize=22)
plt.ylabel('uncertainty / au',fontsize=22)
plt.tick_params(which='both',labelsize=16)

plt.show()





# -*- coding: utf-8 -*-

"""

Uncertainty propagation for a FIR lowpass filter with uncertain cut-off frequency for a rectangular signal.

"""

import matplotlib.pyplot as plt
import numpy as np

import PyDynamic.uncertainty.propagate_MonteCarlo as MC
from PyDynamic.misc.filterstuff import kaiser_lowpass
from PyDynamic.misc.testsignals import rect
from PyDynamic.misc.tools import make_semiposdef
from PyDynamic.uncertainty.propagate_filter import FIRuncFilter

# parameters of simulated measurement
Fs = 100e3		# sampling frequency (in Hz)
Ts = 1 / Fs		# sampling interval length (in s)

# nominal system parameters
fcut = 20e3							# low-pass filter cut-off frequency (6 dB)
L = 100								# filter order
b = kaiser_lowpass(L,fcut,Fs)[0]

# uncertain knowledge: cutoff between 19.5kHz and 20.5kHz
runs = 1000
FC = fcut + (2*np.random.rand(runs)-1)*0.5e3

B = np.zeros((runs,L+1))
for k in range(runs):		# Monte Carlo for filter coefficients of low-pass filter
	B[k,:] = kaiser_lowpass(L,FC[k],Fs)[0]

Ub = make_semiposdef(np.cov(B,rowvar=0))	# covariance matrix of MC result

# simulate input and output signals
time = np.arange(0,499*Ts,Ts)					# time values
noise = 1e-5									# std of white noise
x = rect(time,100*Ts,250*Ts,1.0,noise=noise)	# input signal

y, Uy = FIRuncFilter(x, noise, b, Ub, blow=b)			# apply uncertain FIR filter (GUM formula)
yMC,UyMC = MC.MC(x,noise,b,[1.0],Ub,runs=1000,blow=b)	# apply uncertain FIR filter (Monte Carlo)

plt.figure(1); plt.cla()
plt.plot(time, x, label="input")
plt.plot(time, y, label="output")
plt.xlabel("time / au")
plt.ylabel("signal amplitude / au")
plt.legend()

plt.figure(2);plt.cla()
plt.plot(time, Uy, label="FIR formula")
plt.plot(time, np.sqrt(np.diag(UyMC)), label="Monte Carlo")
plt.xlabel("time / au")
plt.ylabel("signal uncertainty/ au")
plt.legend()
plt.show()



# -*- coding: utf-8 -*-
""" Perform tests on methods to handle DFT and inverse DFT."""

from typing import Callable, Optional, Tuple, Union

import numpy as np
import pytest
from hypothesis import assume, given, strategies as hst
from hypothesis.extra import numpy as hnp
from hypothesis.strategies import composite
from numpy.testing import assert_allclose, assert_almost_equal

from PyDynamic.misc.testsignals import multi_sine
from PyDynamic.uncertainty.propagate_DFT import (
    _apply_window,
    AmpPhase2Time,
    GUM_DFT,
    GUM_iDFT,
    Time2AmpPhase,
)


@pytest.fixture
def multisine_testsignal() -> Tuple[np.ndarray, float]:
    """Additional helper function to create test multi-sine signal"""
    dt = 0.0001
    # set amplitude values of multi-sine components (last parameter is number of
    # components)
    sine_amps = np.random.randint(1, 4, 10)
    # set frequencies of multi-sine components
    sine_freqs = np.linspace(100, 500, len(sine_amps)) * 2 * np.pi
    # define time axis
    time = np.arange(0.0, 0.2, dt)
    time = time[: int(2 ** np.floor(np.log2(len(time))))]
    # measurement noise standard deviation (assume white noise)
    sigma_noise = 0.001
    # generate test signal
    testsignal = multi_sine(time, sine_amps, sine_freqs, noise=sigma_noise)
    return testsignal, sigma_noise


@pytest.fixture(scope="module")
def create_corrmatrix():
    def _create_corrmatrix(rho, Nx, nu=0.5, phi=0.3):
        """ Additional helper function to create a correlation matrix"""
        corrmat = np.zeros((Nx, Nx))
        if rho > 1:
            raise ValueError("Correlation scalar should be less than one.")

        for k in range(1, Nx):
            corrmat += np.diag(np.ones(Nx - k) * rho ** (phi * k ** nu), k)
        corrmat += corrmat.T
        corrmat += np.eye(Nx)

        return corrmat
    return _create_corrmatrix


class TestDFT:
    def test_DFT_iDFT(self, multisine_testsignal):
        """Test GUM_DFT and GUM_iDFT with noise variance as uncertainty"""
        x, ux = multisine_testsignal
        X, UX = GUM_DFT(x, ux ** 2)
        xh, uxh = GUM_iDFT(X, UX)
        assert_almost_equal(np.max(np.abs(x - xh)), 0)
        assert_almost_equal(np.max(ux - np.sqrt(np.diag(uxh))), 0)

    def test_DFT_iDFT_vector(self, multisine_testsignal):
        """Test GUM_DFT and GUM_iDFT with uncertainty vector"""
        x, ux = multisine_testsignal
        ux = (0.1 * x) ** 2
        X, UX = GUM_DFT(x, ux)
        xh, uxh = GUM_iDFT(X, UX)
        assert_almost_equal(np.max(np.abs(x - xh)), 0)
        assert_almost_equal(np.max(np.sqrt(ux) - np.sqrt(np.diag(uxh))), 0)

    def test_DFT_iDFT_fullcov(self, multisine_testsignal, create_corrmatrix):
        """Test GUM_DFT and GUM_iDFT with full covariance matrix"""
        x, ux = multisine_testsignal
        ux = np.ones_like(x) * 0.01 ** 2
        cx = create_corrmatrix(0.95, len(x))
        Ux = np.diag(ux)
        Ux = Ux.dot(cx.dot(Ux))
        X, UX = GUM_DFT(x, Ux)
        xh, Uxh = GUM_iDFT(X, UX)
        assert_almost_equal(np.max(np.abs(x - xh)), 0)
        assert_almost_equal(np.max(Ux - Uxh), 0)

    def test_AmpPhasePropagation(self, multisine_testsignal):
        """Test Time2AmpPhase and AmpPhase2Time with noise variance as uncertainty"""
        testsignal, noise_std = multisine_testsignal
        A, P, UAP = Time2AmpPhase(testsignal, noise_std ** 2)
        x, ux = AmpPhase2Time(A, P, UAP)
        assert_almost_equal(np.max(np.abs(testsignal - x)), 0)


@pytest.fixture
def known_inputs_and_outputs_for_apply_window():
    x = np.array(
        [
            1.81912799e-051,
            -np.inf,
            1.81912799e-051,
            -7.86008596e245,
            1.81912799e-051,
            1.81912799e-051,
            0.00000000e000,
            -9.00719925e015,
            1.81912799e-051,
            1.81912799e-051,
        ]
    )
    Ux = 1.192092896e-07
    window = np.array(
        [
            -1.50000000e000,
            -3.42473686e-168,
            -3.42473686e-168,
            -3.42473686e-168,
            0.00000000e000,
            1.19209290e-007,
            -3.42473686e-168,
            -3.42473686e-168,
            -3.42473686e-168,
            -3.42473686e-168,
        ]
    )
    xw = np.array(
        [
            -2.72869198e-051,
            np.inf,
            -6.23003466e-219,
            2.69187261e078,
            0.00000000e000,
            2.16856955e-058,
            -0.00000000e000,
            3.08472873e-152,
            -6.23003466e-219,
            -6.23003466e-219,
        ]
    )
    Uxw = np.array([2.68220902e-07, 0, 0, 0, 0, 1.69406590e-21, 0, 0, 0, 0])
    return {"inputs": {"x": x, "Ux": Ux, "window": window}, "outputs": (xw, Uxw)}


@composite
def x_Ux_and_window(draw: Callable, ux_type: Optional[Union[np.ndarray, float]] = None):
    """Provide sample x, associated uncertainties and a random window

    Parameters
    ----------
    draw : internal Hypothesis drawing caller (see
        https://hypothesis.readthedocs.io/en/latest/data.html?highlight=composite
        for details)
    ux_type : either np.ndarray or str, optional
        Specification for desired uncertainties as desired return type

    Returns
    -------
    Dictionary to hand over to apply_window(**x_Ux_and_window([...])).
    """
    # Draw dimension of x, window and of Ux in case it is a matrix.
    dim = draw(hst.integers(min_value=1, max_value=20))

    x = draw(hnp.arrays(dtype=float, shape=dim))
    # Prepare drawing Ux as matrix if requested or either as float or matrix else.
    full_ux_strategy = hnp.arrays(dtype=float, shape=(dim, dim))
    float_ux_strategy = hst.floats(min_value=0)
    if ux_type == np.ndarray:
        uncertainty_strategy = full_ux_strategy
    elif ux_type == float:
        uncertainty_strategy = float_ux_strategy
    else:
        uncertainty_strategy = hst.one_of(float_ux_strategy, full_ux_strategy)
    Ux = draw(uncertainty_strategy)
    window = draw(hnp.arrays(dtype=float, shape=dim))

    return {"x": x, "Ux": Ux, "window": window}


@given(x_Ux_and_window())
def test_apply_window(params):
    """Check if application of window to random sample signal works"""
    assert _apply_window(**params)


@given(x_Ux_and_window())
def test_wrong_dimension_x_apply_window(params):
    """Check that wrong dimension of x causes exception"""
    # Make sure x contains more than one value and then break it.
    assume(len(params["x"]) > 1)
    params["x"] = params["x"][1:]

    with pytest.raises(AssertionError):
        _apply_window(**params)


@given(x_Ux_and_window())
def test_wrong_dimension_window_apply_window(params):
    """Check that wrong dimension of x causes exception"""
    # Make sure x contains more than one value and then break it.
    assume(len(params["window"]) > 1)
    params["window"] = params["window"][1:]

    with pytest.raises(AssertionError):
        _apply_window(**params)


@given(x_Ux_and_window(ux_type=np.ndarray))
def test_wrong_dimension_ux_apply_window(params):
    """Check that wrong dimension of x causes exception"""
    # Make sure Ux is at least 2x2 and then break it.
    assume(params["Ux"].shape[0] > 1)

    # Cut Ux by one row. Now ux is not square anymore.
    params["Ux"] = params["Ux"][1:]

    with pytest.raises(AssertionError):
        _apply_window(**params)

    # Cut Ux by one column. Now ux is square again but does not match x.
    params["Ux"] = params["Ux"][:, 1:]

    with pytest.raises(AssertionError):
        _apply_window(**params)


@given(x_Ux_and_window(ux_type=float))
def test_apply_window(params):
    """Check if application of window to random sample with scalar uncertainty works"""
    assert _apply_window(**params)


def test_apply_window_with_known_result(known_inputs_and_outputs_for_apply_window):
    """Test one actually known result for apply_window"""
    assert_allclose(
        actual=_apply_window(**known_inputs_and_outputs_for_apply_window["inputs"]),
        desired=known_inputs_and_outputs_for_apply_window["outputs"],
    )



# -*- coding: utf-8 -*-
"""
This module contains a very first draft for the utilization of the experimental
`Signal` class.

.. note:: The `Signal` class in :mod:`PyDynamic.signals` is experimental and its
signatures might change in the future or it might as well disappear completely. Let
us know if you have special needs or want to make extended use of this feature.
"""

import matplotlib.pyplot as plt
import numpy as np
import scipy.signal as dsp

from PyDynamic import rect
from PyDynamic.signals import Signal


def demonstrate_signal():
    N = 1024
    delta_t = 0.01
    t = np.arange(0, N * delta_t, delta_t)
    x = rect(t, delta_t * N // 4, delta_t * N // 4 * 3)
    ux = 0.02
    signal = Signal(t, x, Ts=delta_t, uncertainty=ux)
    b = dsp.firls(
        15,
        [0, 0.2 * signal.Fs / 2, 0.25 * signal.Fs / 2, signal.Fs / 2],
        [1, 1, 0, 0],
        nyq=signal.Fs / 2,
    )
    Ub = np.diag(b * 1e-1)
    signal.apply_filter(b, filter_uncertainty=Ub)
    signal.plot_uncertainty()
    plt.show()
    bl, al = dsp.bessel(4, 0.2)
    Ul = np.diag(np.r_[al[1:] * 1e-3, bl * 1e-2] ** 2)
    signal.apply_filter(bl, al, filter_uncertainty=Ul)
    signal.plot_uncertainty(fignr=3)
    plt.show()


if __name__ == "__main__":
    demonstrate_signal()



# -*- coding: utf-8 -*-
"""
.. deprecated:: 1.2.71
    The module *deconvolution* will be combined with the module *identification* and
    renamed to *model_estimation* in the next major release 2.0.0. From then on you
    should only use the new module :doc:`PyDynamic.model_estimation` instead. The
    functions :func:`LSFIR`, :func:`LSFIR_unc`, :func:`LSIIR`, :func:`LSIIR_unc`,
    :func:`LSFIR_uncMC` are then prefixed with an "inv" for "inverse", indicating the
    treatment of the reciprocal of frequency response values. Please use the new
    function names (e.g. :func:`PyDynamic.model_estimation.fit_filter.invLSIIR_unc`)
    starting from version 1.4.1. The old function names without preceding "inv" will
    only be preserved until the release prior to version 2.0.0.

The :mod:`PyDynamic.deconvolution.fit_filter` module implements methods for the
design of digital deconvolution filters by least-squares fitting to the reciprocal of
a given frequency response with associated uncertainties.

This module for now still contains the following functions:

* :func:`LSFIR`: Least-squares fit of a digital FIR filter to the reciprocal of a
  given frequency response.
* :func:`LSFIR_unc`: Design of FIR filter as fit to reciprocal of frequency response
  values with uncertainty
* :func:`LSFIR_uncMC`: Design of FIR filter as fit to reciprocal of frequency
  response values with uncertainty via Monte Carlo
* :func:`LSIIR`: Design of a stable IIR filter as fit to reciprocal of frequency
  response values
* :func:`LSIIR_unc`: Design of a stable IIR filter as fit to reciprocal of frequency
  response values with uncertainty
"""

import warnings

from ..model_estimation.fit_filter import (
    invLSFIR,
    invLSFIR_unc,
    invLSFIR_uncMC,
    invLSIIR,
    invLSIIR_unc,
)

__all__ = [
    "LSFIR",
    "LSFIR_unc",
    "LSIIR",
    "LSIIR_unc",
    "LSFIR_uncMC",
]

warnings.warn(
    "The module *deconvolution* will be combined with the module *identification* and"
    "renamed to *model_estimation* in the next major release 2.0.0. From then on you"
    "should only use the new module *model_estimation* instead. The functions"
    "'LSFIR()', 'LSFIR_unc()', 'LSIIR()', 'LSIIR_unc()', 'LSFIR_uncMC()' are then"
    "prefixed with an 'inv' for 'inverse', indicating the treatment of the reciprocal"
    "of frequency response values. Please use the new function names (e.g. "
    ":func:`PyDynamic.model_estimation.fit_filter.invLSIIR_unc`) starting from version"
    "1.4.1.",
    DeprecationWarning,
)


def LSFIR(H, N, tau, f, Fs, Wt=None):
    """.. deprecated:: 1.4.1 Please use :func:`PyDynamic.model_estimation.invLSFIR`
    """
    return invLSFIR(H=H, N=N, tau=tau, f=f, Fs=Fs, Wt=Wt)


def LSFIR_unc(H, UH, N, tau, f, Fs, wt=None, verbose=True, trunc_svd_tol=None):
    """.. deprecated:: 1.4.1 Please use :func:`PyDynamic.model_estimation.invLSFIR_unc`
    """
    return invLSFIR_unc(
        H=H,
        UH=UH,
        N=N,
        tau=tau,
        f=f,
        Fs=Fs,
        wt=wt,
        verbose=verbose,
        trunc_svd_tol=trunc_svd_tol,
    )


def LSFIR_uncMC(H, UH, N, tau, f, Fs, verbose=True):
    """
    .. deprecated:: 1.4.1
        Please use :func:`PyDynamic.model_estimation.invLSFIR_uncMC`
    """
    return invLSFIR_uncMC(H=H, UH=UH, N=N, tau=tau, f=f, Fs=Fs, verbose=verbose)


def LSIIR(Hvals, Nb, Na, f, Fs, tau, justFit=False, verbose=True):
    """.. deprecated:: 1.4.1 Please use :func:`PyDynamic.model_estimation.invLSIIR`"""
    return invLSIIR(
        Hvals=Hvals, Nb=Nb, Na=Na, f=f, Fs=Fs, tau=tau, justFit=justFit, verbose=verbose
    )


def LSIIR_unc(H, UH, Nb, Na, f, Fs, tau=0):
    """.. deprecated:: 1.4.1 Please use :func:`PyDynamic.model_estimation.invLSIIR_unc`
    """
    return invLSIIR_unc(H=H, UH=UH, Nb=Nb, Na=Na, f=f, Fs=Fs, tau=tau)



# -*- coding: utf-8 -*-
"""The module :mod:`PyDynamic.model_estimation.fit_transfer` contains a function
for the identification of transfer function models.

This module contains the following function:

* :func:`fit_som`: Fit second-order model to complex-valued frequency response

"""
import numpy as np

__all__ = ["fit_som"]


def fit_som(f, H, UH=None, weighting=None, MCruns=None, scaling=1e-3):
    """Fit second-order model to complex-valued frequency response

    Fit second-order model (spring-damper model) with parameters
    :math:`S_0, delta` and :math:`f_0` to complex-valued frequency response
    with uncertainty associated with real and imaginary parts.

    For a transformation of an uncertainty associated with amplitude and
    phase to one associated with real and imaginary parts,
    see :mod:`PyDynamic.uncertainty.propagate_DFT.AmpPhase2DFT`.

    Parameters
    ----------
        f: np.ndarray of shape (M,)
            vector of frequencies
        H: np.ndarray of shape (2M,)
            real and imaginary parts of measured frequency response values at
            frequencies f
        UH: np.ndarray of shape (2M,) or (2M,2M)
            uncertainties associated with real and imaginary parts
            When UH is one-dimensional, it is assumed to contain standard
            uncertainties; otherwise it
            is taken as covariance matrix. When UH is not specified no
            uncertainties assoc. with the fit are calculated.
        weighting: str or array
            Type of weighting (None, 'diag', 'cov') or array of weights (
            length two times of f)
        MCruns: int, optional
            Number of Monte Carlo trials for propagation of uncertainties.
            When MCruns is 'None', matrix multiplication
            is used for the propagation of uncertainties. However, in some
            cases this can cause trouble.
        scaling: float
            scaling of least-squares design matrix for improved fit quality
    Returns
    -------
        p: np.ndarray
            vector of estimated model parameters [S0, delta, f0]
        Up: np.ndarray
            covariance associated with parameter estimate
    """
    assert 2 * len(f) == len(H)
    Hr = H[: len(f)]
    Hi = H[len(f) :]

    if isinstance(UH, np.ndarray):
        assert UH.shape[0] == 2 * len(f)
        if len(UH.shape) == 2:
            assert UH.shape[0] == UH.shape[1]

        # propagate to real and imaginary parts of reciprocal using Monte Carlo

        if isinstance(MCruns, int) or isinstance(MCruns, float):
            runs = int(MCruns)
        else:
            runs = 10000
        if len(UH.shape) == 1:
            HR = np.tile(Hr, (runs, 1)) + np.random.randn(runs, len(f)) * np.tile(
                UH[: len(f)], (runs, 1)
            )
            HI = np.tile(Hi, (runs, 1)) + np.random.randn(runs, len(f)) * np.tile(
                UH[len(f) :], (runs, 1)
            )
            HMC = HR + 1j * HI
        else:
            HRI = np.random.multivariate_normal(H, UH, runs)
            HMC = HRI[:, : len(f)] + 1j * HRI[:, len(f) :]

        iRI = np.c_[np.real(1 / HMC), np.imag(1 / HMC)]
        iURI = np.cov(iRI, rowvar=False)

    if isinstance(weighting, str):
        if weighting == "diag":
            W = np.diag(np.diag(iURI))
        elif weighting == "cov":
            W = iURI
        else:
            print("Warning: Specified wrong type of weighting.")
            W = np.eye(2 * len(f))
    elif isinstance(weighting, np.ndarray):
        assert len(weighting) == 2 * len(f)
        W = np.diag(weighting)
    else:
        W = np.eye(2 * len(f))

    if isinstance(UH, np.ndarray):
        # Apply GUM S2
        if isinstance(MCruns, int) or isinstance(MCruns, float):
            # Monte Carlo
            runs = int(MCruns)
            MU = np.zeros((runs, 3))
            for k in range(runs):
                iri = iRI[k, :]
                n = len(f)
                om = 2 * np.pi * f * scaling
                E = np.c_[np.ones(n), 2j * om, -(om ** 2)]
                X = np.r_[np.real(E), np.imag(E)]

                XVX = X.T.dot(np.linalg.solve(W, X))
                XVy = X.T.dot(np.linalg.solve(W, iri))

                MU[k, :] = np.linalg.solve(XVX, XVy)
            MU[:, 1] *= scaling
            MU[:, 2] *= scaling ** 2

            # Calculate S0, delta and f0
            PARS = np.c_[
                1 / MU[:, 0],
                MU[:, 1] / np.sqrt(np.abs(MU[:, 0] * MU[:, 2])),
                np.sqrt(np.abs(MU[:, 0] / MU[:, 2])) / 2 / np.pi,
            ]

            pars = PARS.mean(axis=0)
            Upars = np.cov(PARS, rowvar=False)
        else:  # apply GUM S2 linear propagation
            Hc = Hr + 1j * Hi
            assert np.min(
                np.abs(Hc) > 0
            ), "Frequency response cannot be equal to zero for inversion."
            iri = np.r_[np.real(1 / Hc), np.imag(1 / Hc)]
            n = len(f)
            om = 2 * np.pi * f
            E = np.c_[np.ones(n), 2j * om, -(om ** 2)]
            X = np.r_[np.real(E), np.imag(E)]

            XVX = X.T.dot(np.linalg.solve(W, X))
            XVy = X.T.dot(np.linalg.solve(W, iri))

            mu = np.linalg.solve(XVX, XVy)
            iXVX = np.linalg.inv(XVX)
            XVUVX = np.dot(X.T.dot(np.linalg.solve(W, iURI)), np.linalg.solve(W, X))
            Umu = iXVX.dot(XVUVX).dot(iXVX)

            pars = np.r_[
                1 / mu[0],
                mu[1] / np.sqrt(np.abs(mu[0] * mu[2])),
                np.sqrt(np.abs(mu[0] / mu[2])) / 2 / np.pi,
            ]
            C = np.array(
                [
                    [-1 / mu[0] ** 2, 0, 0],
                    [
                        -mu[1] / (2 * (np.abs(mu[0] + mu[2])) ** (3 / 2)),
                        1 / np.sqrt(np.abs(mu[0] + mu[2])),
                        -mu[1] / (2 * np.abs(mu[0] + mu[2]) ** (3 / 2)),
                    ],
                    [
                        1 / (4 * np.pi * mu[2] * np.sqrt(np.abs(mu[0] / mu[2]))),
                        0,
                        -mu[0]
                        / (4 * np.pi * mu[2] ** 2 * np.sqrt(np.abs(mu[0] / mu[2]))),
                    ],
                ]
            )
            Upars = C.dot(Umu.dot(C.T))

        return pars, Upars
    else:
        Hc = Hr + 1j * Hi
        assert (
            np.min(np.abs(Hc)) > 0
        ), "Frequency response cannot be equal to zero for inversion."
        iri = np.r_[np.real(1 / Hc), np.imag(1 / Hc)]
        n = len(f)
        om = 2 * np.pi * f * scaling
        E = np.c_[np.ones(n), 2j * om, -(om ** 2)]
        X = np.r_[np.real(E), np.imag(E)]

        XVX = X.T.dot(np.linalg.solve(W, X))
        XVy = X.T.dot(np.linalg.solve(W, iri))

        mu = np.linalg.solve(XVX, XVy)
        mu[1] *= scaling
        mu[2] *= scaling ** 2
        pars = np.r_[
            1 / mu[0],
            mu[1] / np.sqrt(np.abs(mu[0] * mu[2])),
            np.sqrt(np.abs(mu[0] / mu[2])) / 2 / np.pi,
        ]
        return pars



# -*- coding: utf-8 -*-
""" Perform tests on the method *uncertainty.propagate_MonteCarlo*"""

import numpy as np
from pytest import raises
import functools
import scipy

from PyDynamic.misc.testsignals import rect
from PyDynamic.misc.tools import make_semiposdef
from PyDynamic.misc.filterstuff import kaiser_lowpass
#from PyDynamic.misc.noise import power_law_acf, power_law_noise, white_gaussian, ARMA
from PyDynamic.uncertainty.propagate_MonteCarlo import MC, SMC, UMC, ARMA, UMC_generic, _UMCevaluate

import matplotlib.pyplot as plt




##### some definitions for all tests

# parameters of simulated measurement
Fs = 100e3        # sampling frequency (in Hz)
Ts = 1 / Fs       # sampling interval length (in s)

# nominal system parameters
fcut = 20e3                                 # low-pass filter cut-off frequency (6 dB)
L = 100                                     # filter order
b1 = kaiser_lowpass(L,   fcut,Fs)[0]
b2 = kaiser_lowpass(L-20,fcut,Fs)[0]

# uncertain knowledge: cutoff between 19.5kHz and 20.5kHz
runs = 20
FC = fcut + (2*np.random.rand(runs)-1)*0.5e3

B = np.zeros((runs,L+1))
for k in range(runs):                       # Monte Carlo for filter coefficients of low-pass filter
    B[k,:] = kaiser_lowpass(L,FC[k],Fs)[0]

Ub = make_semiposdef(np.cov(B,rowvar=0))    # covariance matrix of MC result

# simulate input and output signals
nTime = 500
time  = np.arange(nTime)*Ts                 # time values

# different cases
sigma_noise = 1e-5

# input signal + run methods
x = rect(time,100*Ts,250*Ts,1.0,noise=sigma_noise)



##### actual tests

def test_MC(visualizeOutput=False):
    # run method
    y,Uy = MC(x,sigma_noise,b1,[1.0],Ub,runs=runs,blow=b2)

    assert len(y) == len(x)
    assert Uy.shape == (x.size, x.size)

    if visualizeOutput:
        # visualize input and mean of system response
        plt.plot(time, x)
        plt.plot(time, y)

        # visualize uncertainty of output
        plt.plot(time, y - np.sqrt(np.diag(Uy)), linestyle="--", linewidth=1, color="red")
        plt.plot(time, y + np.sqrt(np.diag(Uy)), linestyle="--", linewidth=1, color="red")

        plt.show()


# this does not run through yet
#def test_SMC():
#    # run method
#    y,Uy = SMC(x, sigma_noise, b1, [1.0], Ub, runs=runs)
#
#    assert len(y) == len(x)
#    assert Uy.shape == (x.size, x.size)


def test_UMC(visualizeOutput=False):
    # run method
    y, Uy, p025, p975, happr = UMC(x, b1, [1.0], Ub, blow=b2, sigma=sigma_noise, runs=runs, runs_init=10, nbins=10)

    assert len(y) == len(x)
    assert Uy.shape == (x.size, x.size)
    assert p025.shape[1] == len(x)
    assert p975.shape[1] == len(x)
    assert isinstance(happr, dict)

    if visualizeOutput:
        # visualize input and mean of system response
        plt.plot(time, x)
        plt.plot(time, y)

        # visualize uncertainty of output
        plt.plot(time, y - np.sqrt(np.diag(Uy)), linestyle="--", linewidth=1, color="red")
        plt.plot(time, y + np.sqrt(np.diag(Uy)), linestyle="--", linewidth=1, color="red")

        # visualize central 95%-quantile
        plt.plot(time, p025.T, linestyle=":", linewidth=1, color="gray")
        plt.plot(time, p975.T, linestyle=":", linewidth=1, color="gray")

        # visualize the bin-counts
        key = list(happr.keys())[0]
        for ts, be, bc in zip(time, happr[key]["bin-edges"].T, happr[key]["bin-counts"].T):
            plt.scatter(ts*np.ones_like(bc), be[1:], bc)

        plt.show()


def test_UMC_generic(visualizeOutput=False):

    x_shape = (5,6,7)
    draw_samples = lambda size: np.random.rand(size, *x_shape)
    evaluate = functools.partial(np.mean, axis=1)

    # run UMC
    y, Uy, happr, output_shape = UMC_generic(draw_samples, evaluate, runs=100, blocksize=20, runs_init=10)
    assert y.size == Uy.shape[0]
    assert Uy.shape == (y.size, y.size)
    assert isinstance(happr, dict)
    assert output_shape == (5,7)

    # run without parallel computation
    y, Uy, happr, output_shape = UMC_generic(draw_samples, evaluate, runs=100, blocksize=20, runs_init=10, n_cpu=1)
    assert y.size == Uy.shape[0]
    assert Uy.shape == (y.size, y.size)
    assert isinstance(happr, dict)
    assert output_shape == (5,7)

    # run again, but only return all simulations
    y, Uy, happr, output_shape, sims = UMC_generic(draw_samples, evaluate, runs=100, blocksize=20, runs_init=10, return_samples=True)
    assert y.size == Uy.shape[0]
    assert Uy.shape == (y.size, y.size)
    assert isinstance(happr, dict)
    assert output_shape == (5,7)
    assert isinstance(sims, dict)
    assert sims["samples"][0].shape == x_shape
    assert sims["results"][0].shape == output_shape


def test_compare_MC_UMC():

    np.random.seed(12345)

    y_MC, Uy_MC = MC(x,sigma_noise,b1,[1.0],Ub,runs=2*runs,blow=b2)
    y_UMC, Uy_UMC, _, _, _ = UMC(x, b1, [1.0], Ub, blow=b2, sigma=sigma_noise, runs=2*runs, runs_init=10)

    # both methods should yield roughly the same results
    assert np.allclose(y_MC, y_UMC, atol=5e-4)
    assert np.allclose(Uy_MC, Uy_UMC, atol=5e-4)


def test_noise_ARMA():
    length = 100
    phi = [1/3, 1/4, 1/5]
    theta = [1, -1 ]

    e = ARMA(length, phi = phi, theta = theta)

    assert len(e) == length



# -*- coding: utf-8 -*-
""" Perform tests on methods specialized for second order dynamic systems."""

from matplotlib.pyplot import np as np
from pytest import approx
from scipy.signal import freqs

from PyDynamic.misc.SecondOrderSystem import sos_FreqResp, sos_phys2filter, \
    sos_realimag, sos_absphase

np.random.seed(12345)

Fs = 100e3
delta = 0.0001
f0 = float(Fs / 4 + np.abs(np.random.randn(1)) * Fs / 8)
S0 = 1.0
fe5 = np.linspace(0, Fs / 2, 100000)
K = 100
fe3 = np.linspace(0, Fs / 2, 1000)
udelta = 1e-12 * delta
uf0 = 1e-12 * f0
uS0 = 1e-12 * S0


def test_sos_freqresp():
    H = sos_FreqResp(S0, delta, f0, fe5)
    indmax = np.abs(H).argmax()
    assert np.round(np.abs(f0 - fe5[indmax])) <= 0.01 * f0
    assert np.abs(H[0]) == approx(S0)
    Hmulti = sos_FreqResp(
        S0 * np.ones(K), delta * np.ones(K), f0 * np.ones(K), fe5)
    assert Hmulti.shape[1] == K


def test_sos_phys2filter():
    b, a = sos_phys2filter(S0, delta, f0)
    H = freqs(b, a, 2 * np.pi * fe5)[1]
    indmax = np.abs(H).argmax()
    assert np.round(np.abs(f0 - fe5[indmax])) <= 0.01 * f0
    assert np.abs(H[0]) == approx(S0)
    bmulti, amulti = sos_phys2filter(
        S0 * np.ones(K), delta * np.ones(K), f0 * np.ones(K))
    assert len(bmulti[0]) == K
    assert amulti.shape == (K, 3)


def test_sos_realimag():
    Hmean, Hcov = sos_realimag(S0, delta, f0, uS0, udelta, uf0, fe3, runs=100)
    assert Hcov.shape, (2 * len(fe3) == 2 * len(fe3))
    H = sos_FreqResp(S0, delta, f0, fe3)
    assert np.linalg.norm(H) == approx(np.linalg.norm(Hmean))


def test_sos_absphase():
    Hmean, Hcov = sos_absphase(S0, delta, f0, uS0, udelta, uf0, fe3, runs=100)
    assert Hcov.shape == (2 * len(fe3), 2 * len(fe3))
    H = sos_FreqResp(S0, delta, f0, fe3)
    assert np.linalg.norm(H) == approx(np.linalg.norm(Hmean))



# -*- coding: utf-8 -*-
"""
This module contains functions for the propagation of uncertainties through
the application of a digital filter using the GUM approach.

This modules contains the following functions:

* :func:`FIRuncFilter`: Uncertainty propagation for signal y and uncertain FIR
  filter theta
* :func:`IIRuncFilter`: Uncertainty propagation for the signal x and the uncertain
  IIR filter (b,a)

.. note:: The Elster-Link paper for FIR filters assumes that the autocovariance
          is known and that noise is stationary!

"""

import numpy as np
from scipy.linalg import toeplitz
from scipy.signal import lfilter, lfilter_zi, dimpulse
from scipy.signal import convolve
from ..misc.tools import trimOrPad

__all__ = ["FIRuncFilter", "IIRuncFilter"]


def _fir_filter(x, theta, Ux=None, Utheta=None, initial_conditions="constant"):
    """Uncertainty propagation for signal x with covariance Ux
       and uncertain FIR filter theta with covariance Utheta.

       If either Ux or Utheta are omitted (None), then corresponding terms are not
       calculated to reduce computation time.

    Parameters
    ----------
    x : np.ndarray
        filter input signal
    theta : np.ndarray
        FIR filter coefficients
    Ux : np.ndarray, optional
        covariance matrix associated with x
        if the signal is fully certain, use `Ux = None` (default) to make use of more efficient calculations.
    Utheta : np.ndarray, optional
        covariance matrix associated with theta
        if the filter is fully certain, use `Utheta = None` (default) to make use of more efficient calculations.
        see also the comparison given in <examples\Digital filtering\FIRuncFilter_runtime_comparison.py>
    initial_conditions : str, optional
        constant: assume signal + uncertainty are constant before t=0 (default)
        zero: assume signal + uncertainty are zero before t=0


    Returns
    -------
    y : np.ndarray
        FIR filter output signal
    Uy : np.ndarray
        covariance matrix of filter output y


    References
    ----------
    * Elster and Link 2008 [Elster2008]_

    .. seealso:: :mod:`PyDynamic.model_estimation.fit_filter`

    """

    # Note to future developers:
    # The functions _fir_filter and _fir_filter_diag share
    # the same logic. If adjustments become necessary (e.g.
    # due to bug fixing) please also consider adjusting it
    # in the other function as well.

    Ntheta = len(theta)  # FIR filter size

    if initial_conditions == "constant":
        x0 = x[0]

    # Note: currently only used in testing for comparison against Monte Carlo method
    elif initial_conditions == "zero":
        x0 = 0.0

    else:
        raise ValueError(
            f"_fit_filter: You provided 'initial_conditions' = '{initial_conditions}'."
            f"However, only 'zero' or 'constant' are currently supported."
        )

    # propagate filter
    y, _ = lfilter(theta, 1.0, x, zi=x0 * lfilter_zi(theta, 1.0))

    # propagate uncertainty
    Uy = np.zeros((len(x), len(x)))

    ## only calculate subterms, that are non-zero (compare eq. 34 of paper)
    if Ux is not None:
        ## extend covariance Ntheta steps into the past
        if initial_conditions == "constant":
            Ux_extended = _stationary_prepend_covariance(Ux, Ntheta - 1)

        elif initial_conditions == "zero":
            # extend covariance Ntheta steps into the past
            Ux_extended = np.pad(
                Ux,
                ((Ntheta - 1, 0), (Ntheta - 1, 0)),
                "constant",
                constant_values=0,
            )

        # calc subterm theta^T * Ux * theta
        Uy += convolve(np.outer(theta, theta), Ux_extended, mode="valid")

    if Utheta is not None:
        ## extend signal Ntheta steps into the past
        x_extended = np.r_[np.full((Ntheta - 1), x0), x]

        # calc subterm x^T * Utheta * x
        Uy += convolve(np.outer(x_extended, x_extended), Utheta, mode="valid")

    if (Ux is not None) and (Utheta is not None):
        # calc subterm Tr(Ux * Utheta)
        Uy += convolve(Ux_extended, Utheta.T, mode="valid")

    return y, Uy


def _fir_filter_diag(
    x, theta, Ux_diag=None, Utheta_diag=None, initial_conditions="constant"
):
    """Uncertainty propagation for signal x with covariance diagonal Ux_diag
       and uncertain FIR filter theta with covariance diagonal Utheta_diag.

       If either Ux_diag or Utheta_diag are omitted (None), then corresponding terms are not
       calculated to reduce computation time.

    Parameters
    ----------
    x : np.ndarray
        filter input signal
    theta : np.ndarray
        FIR filter coefficients
    Ux_diag : np.ndarray, optional
        diagonal of covariance matrix ([ux11^2, ux22^2, ..., uxnn^2])
        if the signal is fully certain, use `Ux_diag = None` (default) to make use of more efficient calculations.
    Utheta_diag : np.ndarray, optional
        diagonal of covariance matrix ([ut11^2, ut22^2, ..., utnn^2])
        if the filter is fully certain, use `Utheta_diag = None` (default) to make use of more efficient calculations.
        see also the comparison given in <examples\Digital filtering\FIRuncFilter_runtime_comparison.py>
    initial_conditions : str, optional
        constant: assume signal + uncertainty are constant before t=0 (default)
        zero: assume signal + uncertainty are zero before t=0


    Returns
    -------
    y : np.ndarray
        FIR filter output signal
    Uy_diag : np.ndarray
        diagonal of covariance matrix of filter output y


    References
    ----------
    * Elster and Link 2008 [Elster2008]_

    .. seealso:: :mod:`PyDynamic.model_estimation.fit_filter`

    """

    # Note to future developers:
    # The functions _fir_filter and _fir_filter_diag share
    # the same logic. If adjustments become necessary (e.g.
    # due to bug fixing) please also consider adjusting it
    # in the other function as well.

    Ntheta = len(theta)  # FIR filter size

    if initial_conditions == "constant":
        x0 = x[0]

    # Note: currently only used in testing for comparison against Monte Carlo method
    elif initial_conditions == "zero":
        x0 = 0.0

    else:
        raise ValueError(
            f"_fit_filter: You provided 'initial_conditions' = '{initial_conditions}'."
            f"However, only 'zero' or 'constant' are currently supported."
        )

    # propagate filter
    y, _ = lfilter(theta, 1.0, x, zi=x0 * lfilter_zi(theta, 1.0))

    # propagate uncertainty
    Uy_diag = np.zeros(len(x))

    ## only calculate subterms, that are non-zero (compare eq. 34 of paper)
    if Ux_diag is not None:
        ## extend covariance Ntheta steps into the past
        if initial_conditions == "constant":
            Ux0 = Ux_diag[0]
        elif initial_conditions == "zero":
            Ux0 = 0.0
        Ux_diag_extended = np.r_[np.full((Ntheta - 1), Ux0), Ux_diag]

        # calc subterm theta^T * Ux * theta
        Uy_diag += convolve(np.square(theta), Ux_diag_extended, mode="valid")

    if Utheta_diag is not None:
        ## extend signal Ntheta steps into the past
        x_extended = np.r_[np.full((Ntheta - 1), x0), x]

        # calc subterm x^T * Utheta * x
        Uy_diag += convolve(np.square(x_extended), Utheta_diag, mode="valid")

    if (Ux_diag is not None) and (Utheta_diag is not None):
        # calc subterm Tr(Ux * Utheta)
        Uy_diag += convolve(Ux_diag_extended, Utheta_diag, mode="valid")

    return y, Uy_diag


def _stationary_prepend_covariance(U, n):
    """ Prepend covariance matrix U by n steps into the past"""

    c = np.r_[U[:, 0], np.zeros(n)]
    r = np.r_[U[0, :], np.zeros(n)]

    U_adjusted = toeplitz(c, r)
    U_adjusted[n:, n:] = U

    return U_adjusted


def FIRuncFilter(
    y,
    sigma_noise,
    theta,
    Utheta=None,
    shift=0,
    blow=None,
    kind="corr",
    return_full_covariance=False,
):
    """Uncertainty propagation for signal y and uncertain FIR filter theta

    A preceding FIR low-pass filter with coefficients `blow` can be provided optionally.

    This method keeps the signature of `PyDynamic.uncertainty.FIRuncFilter`, but internally
    works differently and can return a full covariance matrix. Also, sigma_noise can be a full
    covariance matrix.

    Parameters
    ----------
    y : np.ndarray
        filter input signal
    sigma_noise : float or np.ndarray
        float:    standard deviation of white noise in y
        1D-array: interpretation depends on kind
        2D-array: full covariance of input
    theta : np.ndarray
        FIR filter coefficients
    Utheta : np.ndarray, optional
        1D-array: coefficient-wise standard uncertainties of filter
        2D-array: covariance matrix associated with theta
        if the filter is fully certain, use `Utheta = None` (default) to make use of more efficient calculations.
        see also the comparison given in <examples\Digital filtering\FIRuncFilter_runtime_comparison.py>
    shift : int, optional
        time delay of filter output signal (in samples) (defaults to 0)
    blow : np.ndarray, optional
        optional FIR low-pass filter
    kind : string
        only meaningful in combination with sigma_noise a 1D numpy array
        "diag": point-wise standard uncertainties of non-stationary white noise
        "corr": single sided autocovariance of stationary (colored/correlated)
        noise (default)
    return_full_covariance : bool, optional
        whether or not to return a full covariance of the output, defaults to False

    Returns
    -------
    x : np.ndarray
        FIR filter output signal
    Ux : np.ndarray
        return_full_covariance == False : point-wise standard uncertainties associated with x (default)
        return_full_covariance == True : covariance matrix containing uncertainties associated with x


    References
    ----------
    * Elster and Link 2008 [Elster2008]_

    .. seealso:: :mod:`PyDynamic.model_estimation.fit_filter`

    """

    # Check for special cases, that we can compute much faster:
    ## These special cases come from the default behavior of the `old` FIRuncFilter
    ## implementation, which always returned only the square-root of the main diagonal
    ## of the covariance matrix.
    ## The special case is activated, if the user does not want a full covariance of
    ## the result (return_full_covariance == False). Furthermore, sigma_noise and Utheta
    ## must be representable by pure diagonal covariance matricies (i.e. they are of type
    ## None, float or 1D-array). By the same reasoning, we need to exclude cases
    ## of preceding low-pass filtering, as the covariance of the low-pass-filtered
    ## signal would no longer be purely diagonal.
    ## Computation is then based on eq. 33 [Elster2008]_ - which is substantially
    ## faster (typically by orders of magnitude) when compared to the full covariance
    ## calculation.
    ## Check this example: <examples\Digital filtering\FIRuncFilter_runtime_comparison.py>
    
    # note to user
    if not return_full_covariance:
        print(
            "FIRuncFilter: Output uncertainty will be given as 1D-array of point-wise "
            "standard uncertainties. Although this requires significantly lesser computations, "
            "it ignores correlation information. Every FIR-filtered signal will have "
            "off-diagonal entries in its covariance matrix (assuming the filter is longer "
            "than 1). To get the full output covariance matrix, use 'return_full_covariance=True'."
        )

    if (
        (not return_full_covariance)  # no need for full covariance
        and (  # cases of sigma_noise, that support the faster computation
            isinstance(sigma_noise, float)
            or (isinstance(sigma_noise, np.ndarray) and len(sigma_noise.shape) == 1)
            or (sigma_noise is None)
        )
        and (  # cases of Utheta, that support the faster computation
            isinstance(Utheta, float)
            or (isinstance(Utheta, np.ndarray) and len(Utheta.shape) == 1)
            or (Utheta is None)
        )
        # the low-pass-filtered signal wouldn't have pure diagonal covariance
        and (blow is None)
        # if sigma_noise is 1D, it must represent the diagonal of a covariance matrix
        and (kind == "diag" or not isinstance(sigma_noise, np.ndarray))
    ):

        if isinstance(sigma_noise, float):
            Uy_diag = np.full_like(y, sigma_noise ** 2)
        elif isinstance(sigma_noise, np.ndarray):
            Uy_diag = np.square(sigma_noise)
        else:
            Uy_diag = sigma_noise

        if isinstance(Utheta, float):
            Utheta_diag = np.full_like(theta, Utheta)
        elif isinstance(Utheta, np.ndarray):
            Utheta_diag = np.square(Utheta)
        else:
            Utheta_diag = Utheta

        x, Ux_diag = _fir_filter_diag(
            y, theta, Uy_diag, Utheta_diag, initial_conditions="constant"
        )
        return x, np.sqrt(np.abs(Ux_diag))

    # otherwise, the method computes full covariance information
    else:
        Ntheta = len(theta)  # FIR filter size

        # check which case of sigma_noise is necessary
        if isinstance(sigma_noise, float):
            Uy = np.diag(np.full(len(y), sigma_noise ** 2))

        elif isinstance(sigma_noise, np.ndarray):

            if len(sigma_noise.shape) == 1:
                if kind == "diag":
                    Uy = np.diag(sigma_noise ** 2)
                elif kind == "corr":
                    Uy = toeplitz(trimOrPad(sigma_noise, len(y)))
                else:
                    raise ValueError("unknown kind of sigma_noise")

            elif len(sigma_noise.shape) == 2:
                Uy = sigma_noise

        else:
            raise ValueError(
                "Unsupported value of sigma_noise. Please check the documentation."
            )

        # filter operation(s)
        if isinstance(blow, np.ndarray):
            # apply (fully certain) lowpass-filter
            xlow, Ulow = _fir_filter(y, blow, Uy, None, initial_conditions="constant")

            # apply filter to lowpass-filtered signal
            x, Ux = _fir_filter(
                xlow, theta, Ulow, Utheta, initial_conditions="constant"
            )

        else:
            # apply filter to input signal
            x, Ux = _fir_filter(y, theta, Uy, Utheta, initial_conditions="constant")

        # shift result
        if shift != 0:
            x = np.roll(x, -int(shift))
            Ux = np.roll(Ux, (-int(shift), -int(shift)))

        if return_full_covariance:
            return x, Ux
        else:
            return x, np.sqrt(np.abs(np.diag(Ux)))


def IIRuncFilter(x, noise, b, a, Uab):
    """
    Uncertainty propagation for the signal x and the uncertain IIR filter (b,a)

    Parameters
    ----------
    x: np.ndarray
        filter input signal
    noise: float
        signal noise standard deviation
    b: np.ndarray
        filter numerator coefficients
    a: np.ndarray
        filter denominator coefficients
    Uab: np.ndarray
        covariance matrix for (a[1:],b)

    Returns
    -------
    y: np.ndarray
        filter output signal
    Uy: np.ndarray
        uncertainty associated with y

    References
    ----------
        * Link and Elster [Link2009]_

    """

    if not isinstance(noise, np.ndarray):
        noise = noise * np.ones_like(x)  # translate iid noise to vector

    p = len(a) - 1

    # Adjust dimension for later use.
    if not len(b) == len(a):
        b = np.hstack((b, np.zeros((len(a) - len(b),))))

    # From discrete-time transfer function to state space representation.
    [A, bs, c, b0] = tf2ss(b, a)

    A = np.matrix(A)
    bs = np.matrix(bs)
    c = np.matrix(c)

    phi = np.zeros((2 * p + 1, 1))
    dz = np.zeros((p, p))
    dz1 = np.zeros((p, p))
    z = np.zeros((p, 1))
    P = np.zeros((p, p))

    y = np.zeros((len(x),))
    Uy = np.zeros((len(x),))

    Aabl = np.zeros((p, p, p))
    for k in range(p):
        Aabl[0, k, k] = -1

    # implementation of the state-space formulas from the paper
    for n in range(len(y)):
        for k in range(p):  # derivative w.r.t. a_1,...,a_p
            dz1[:, k] = A * dz[:, k] + np.squeeze(Aabl[:, :, k]) * z
            phi[k] = c * dz[:, k] - b0 * z[k]
        phi[p + 1] = -np.matrix(a[1:]) * z + x[n]  # derivative w.r.t. b_0
        for k in range(p + 2, 2 * p + 1):  # derivative w.r.t. b_1,...,b_p
            phi[k] = z[k - (p + 1)]
        P = A * P * A.T + noise[n] ** 2 * (bs * bs.T)
        y[n] = c * z + b0 * x[n]
        Uy[n] = phi.T * Uab * phi + c * P * c.T + b[0] ** 2 * noise[n] ** 2
        z = A * z + bs * x[n]  # update of the state equations
        dz = dz1

    Uy = np.sqrt(np.abs(Uy))  # calculate point-wise standard uncertainties

    return y, Uy



# -*- coding: utf-8 -*-
"""Install PyDynamic in Python path and provide all packaging metadata."""

from os import path

from setuptools import find_packages, setup

# Get release version from PyDynamic __init__.py
from PyDynamic import __version__ as VERSION


def get_readme():
    """Get README.md's content"""
    this_directory = path.abspath(path.dirname(__file__))
    with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
        return f.read()


setup(
    metadata_version="2.1",
    name="PyDynamic",
    version=VERSION,
    description="A software package for the analysis of dynamic measurements",
    long_description=get_readme(),
    long_description_content_type="text/markdown",
    url="https://ptb-pst1.github.io/PyDynamic/",
    download_url="https://github.com/PTB-PSt1/PyDynamic/releases/download/v{0}/"
    "PyDynamic-{0}.tar.gz".format(VERSION),
    author=u"Sascha Eichstädt, Maximilian Gruber, Björn Ludwig, Thomas Bruns, "
    "Martin Weber",
    author_email="sascha.eichstaedt@ptb.de",
    keywords="uncertainty dynamic deconvolution metrology",
    packages=find_packages(exclude=["test"]),
    project_urls={
        "Documentation": "https://pydynamic.readthedocs.io/en/v{}/".format(VERSION),
        "Source": "https://github.com/PTB-PSt1/PyDynamic/tree/v{}/".format(VERSION),
        "Tracker": "https://github.com/PTB-PSt1/PyDynamic/issues",
    },
    install_requires=["matplotlib", "numpy", "pandas", "scipy"],
    # This allow to do "pip install PyDynamic[examples]" and get the dependencies to
    # execute the Jupyter Notebook examples.
    extras_require={"examples": ["notebook"],},
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Utilities",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or "
        "later (LGPLv3+)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
        "Typing :: Typed",
    ],
)



# -*- coding: utf-8 -*-
r"""
The :mod:`PyDynamic.misc.filterstuff` module is a collection of methods which are
related to filter design.

This module contains the following functions:

* :func:`db`: Calculation of decibel values :math:`20\log_{10}(x)` for a vector of
  values
* :func:`ua`: Shortcut for calculation of unwrapped angle of complex values
* :func:`grpdelay`: Calculation of the group delay of a digital filter
* :func:`mapinside`: Maps the roots of polynomial with coefficients :math:`a`
  to the unit circle
* :func:`kaiser_lowpass`: Design of a FIR lowpass filter using the window technique
  with a Kaiser window.
* :func:`isstable`: Determine whether a given IIR filter is stable
* :func:`savitzky_golay`: Smooth (and optionally differentiate) data with a
  Savitzky-Golay filter

"""

import numpy as np

__all__ = [
    "db",
    "ua",
    "grpdelay",
    "mapinside",
    "kaiser_lowpass",
    "isstable",
    "savitzky_golay",
]


def db(vals):
    """Calculation of decibel values :math:`20\log_{10}(x)` for a vector of values"""
    return 20 * np.log10(np.abs(vals))


def ua(vals):
    """Shortcut for calculation of unwrapped angle of complex values"""
    return np.unwrap(np.angle(vals))


def grpdelay(b, a, Fs, nfft=512):
    """Calculation of the group delay of a digital filter
   
    Parameters
    ----------
        b: ndarray
            IIR filter numerator coefficients
        a: ndarray
            IIR filter denominator coefficients
        Fs: float
            sampling frequency of the filter
        nfft: int
            number of FFT bins

    Returns
    -------
        group_delay: np.ndarray
            group delay values
        frequencies: ndarray
            frequencies at which the group delay is calculated

    References
    ----------
    * Smith, online book [Smith]_

    """

    Na = len(a) - 1
    Nb = len(b) - 1

    c = np.convolve(b, a[::-1])  # c(z) = b(z)*a(1/z)*z^(-oa)
    cr = c * np.arange(Na + Nb + 1)  # derivative of c wrt 1/z
    num = np.fft.fft(cr, 2 * nfft)
    den = np.fft.fft(c, 2 * nfft)
    tol = 1e-12

    polebins = np.nonzero(abs(den) < tol)
    for p in polebins:
        num[p] = 0.0
        den[p] = 1.0

    gd = np.real(num / den) - Na

    f = np.arange(0.0, 2 * nfft - 1) / (2 * nfft) * Fs
    f = f[: nfft + 1]
    gd = gd[: len(f)]
    return gd, f


def mapinside(a):
    """ Maps the roots of polynomial to the unit circle.

    Maps the roots of polynomial with coefficients :math:`a` to the unit circle.
    
    Parameters
    ----------
        a: ndarray
           polynomial coefficients
    Returns
    -------
        a: ndarray
           polynomial coefficients with all roots inside or on the unit circle
    """
    v = np.roots(a)
    inds = np.nonzero(abs(v) > 1)
    v[inds] = 1 / np.conj(v[inds])
    return np.poly(v)


def kaiser_lowpass(L, fcut, Fs, beta=8.0):
    """ Design of a FIR lowpass filter using the window technique with a
    Kaiser window.

    This method uses a Kaiser window. Filters of that type are often used as
    FIR low-pass filters due to their linear phase.
    
    Parameters
    ----------
        L: int
           filter order (window length)
        fcut: float
              desired cut-off frequency
        Fs: float
            sampling frequency
        beta: float
              scaling parameter for the Kaiser window
    Returns
    -------
        blow: ndarray
              FIR filter coefficients
        shift: int
               delay of the filter (in samples)
    
    """
    from scipy.signal import firwin

    if np.mod(L, 2) == 0:
        L = L + 1
    blow = firwin(L, 2 * fcut / Fs, window=("kaiser", beta))
    shift = L / 2
    return blow, shift


def isstable(b, a, ftype="digital"):
    """Determine whether `IIR filter (b,a)` is stable

    Determine whether `IIR filter (b,a)` is stable by checking roots of the
    polynomial ´a´.
    
    Parameters
    ----------
        b: ndarray
            filter numerator coefficients
        a: ndarray
            filter denominator coefficients
        ftype: string
            type of filter (`digital` or `analog`)
    Returns
    -------
        stable: bool
            whether filter is stable or not
                
    """
    v = np.roots(a)
    if ftype == "digital":
        return not np.any(np.abs(v) > 1.0)
    elif ftype == "analog":
        return not np.any(np.real(v) < 0)
    else:
        raise ValueError(f"isstable: filter type 'ftype={ftype}'unknown. Please "
                         f"check documentation and choose a valid assignment for "
                         f"'ftype' or leave out to fallback to default.")


def savitzky_golay(y, window_size, order, deriv=0, delta=1.0):
    """Smooth (and optionally differentiate) data with a Savitzky-Golay filter

    The Savitzky-Golay filter removes high frequency noise from data.
    It has the advantage of preserving the original shape and
    features of the signal better than other types of filtering
    approaches, such as moving averages techniques.

    Source obtained from scipy cookbook (online), downloaded 2013-09-13

    Parameters
    ----------
        y: ndarray, shape (N,)
           the values of the time history of the signal
        window_size: int
           the length of the window. Must be an odd integer number
        order: int
           the order of the polynomial used in the filtering. Must be less
           then `window_size` - 1.
        deriv: int, optional
            The order of the derivative to compute.  This must be a
            nonnegative integer.  The default is 0, which means to filter
            the data without differentiating.
        delta: float, optional
            The spacing of the samples to which the filter will be applied.
            This is only used if deriv > 0. This includes a factor
            :math:`n! / h^n`, where :math:`n` is represented by `deriv` and
            :math:`1/h` by `delta`.

    Returns
    -------
         ys: ndarray, shape (N,)
            the smoothed signal (or it's n-th derivative).

    Notes
    -----
    The Savitzky-Golay is a type of low-pass filter, particularly
    suited for smoothing noisy data. The main idea behind this
    approach is to make for each point a least-square fit with a
    polynomial of high order over a odd-sized window centered at
    the point.


    References
    ----------
    * Savitzky et al. [Savitzky]_
    * Numerical Recipes [NumRec]_

    """
    from math import factorial

    try:
        window_size = np.abs(np.int(window_size))
        order = np.abs(np.int(order))
    except ValueError:
        raise ValueError("window_size and order have to be of type int")
    if window_size % 2 != 1 or window_size < 1:
        raise TypeError("window_size size must be a positive odd number")
    if window_size < order + 2:
        raise TypeError("window_size is too small for the polynomials order")
    order_range = range(order + 1)
    half_window = (window_size - 1) // 2
    # precompute coefficients
    b = np.mat(
        [[k ** i for i in order_range] for k in range(-half_window, half_window + 1)]
    )
    m = np.linalg.pinv(b).A[deriv] * factorial(deriv) / delta ** deriv
    # pad the signal at the extremes with
    # values taken from the signal itself
    firstvals = y[0] - np.abs(y[1 : half_window + 1][::-1] - y[0])
    lastvals = y[-1] + np.abs(y[-half_window - 1 : -1][::-1] - y[-1])
    y = np.concatenate((firstvals, y, lastvals))
    return np.convolve(m[::-1], y, mode="valid")



# -*- coding: utf-8 -*-
"""
The package :doc:`PyDynamic.model_estimation` implements methods for the model
estimation by least-squares fitting and the identification of transfer function models
with associated uncertainties.
"""

# See http://mathmet.org/projects/14SIP08 and
# https://www.github.com/PTB-PSt1/PyDynamic

from .fit_filter import (LSFIR, LSIIR, invLSFIR, invLSFIR_unc, invLSFIR_uncMC,
                         invLSIIR, invLSIIR_unc)
from .fit_transfer import fit_som

__all__ = [
    "LSFIR",
    "LSIIR",
    "invLSFIR",
    "invLSIIR",
    "invLSFIR_unc",
    "invLSFIR_uncMC",
    "invLSIIR_unc",
    "fit_som",
]



"""Perform a speed comparison for uncertainty.propagate_filter.FIRuncFilter
 with different value-types of the covariance of theta
"""

import numpy as np
import time
import pandas as pd

from PyDynamic.uncertainty.propagate_filter import FIRuncFilter

def evaluate(signal, filter, case="full/zero/none"):
    # different covariance matrices
    if case == "full":
        U_theta = 1e-2 * np.eye(filter_length)
    elif case == "zero":
        U_theta = np.zeros((filter_length, filter_length))
    elif case == "none":
        U_theta = None
    else:
        raise NotImplementedError
    
    # measure runtime (timeit.timeit is unable to use both local+global variables)
    start = time.time()
    y, Uy = FIRuncFilter(signal, sigma_noise, theta, U_theta, blow=None)
    end = time.time()
    runtime = end - start

    return runtime, y, Uy


filter_lengths = [10, 50, 100, 200, 500, 1000]
cases = ["full", "zero", "none"]
results = pd.DataFrame(index=filter_lengths, columns=cases)


# prepare signal
signal_length = 2000
sigma_noise = 1e-2  # 1e-5
signal = sigma_noise * np.random.randn(signal_length)

# print table header
table_hdr = "{0:<15s} | {1:<15s} | {2:<10s} | {3:<15s}"
table_fmt = "{0:<15d} | {1:<15d} | {2:<10s} | {3}"
print(table_hdr.format("signal-length", "filter-length", "case", "uy[:2]"))
print("-"*68)

for filter_length in filter_lengths:

    # prepare filter parameters
    theta = np.random.random(filter_length)
    
    for case in cases:
        runtime, y, uy = evaluate(signal, theta, case)
        results.loc[filter_length, case] = runtime
        print(table_fmt.format(signal_length, filter_length, case, uy[:2]))
    print("-"*68)

# The expected result is, that `full` and `zero` should be roughly as fast (same range/magnitude).
# `none` will be much faster and yields the same results as `zero`.
print("\n\nThese are the observed execution times (in seconds) for \ndifferent combinations of filter-lengths and `case`:")
print(results)
